# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import datetime
import os
import sys

import pytz

sys.path.insert(0, os.path.abspath(".."))

import btrack  # noqa: E402

# -- Project information -----------------------------------------------------

project = "Bayesian Tracker (btrack) 🔬💻"
author = "Alan R Lowe"
year = datetime.datetime.now(tz=pytz.timezone("GMT")).year
copyright = f"2014-{year}, {author}"  # noqa: A001


# The full version, including alpha/beta/rc tags
version = f"v{btrack.__version__}"
release = version


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx_panels",
    "sphinx_automodapi.automodapi",
    "numpydoc",
    "sphinx_rtd_theme",
]

numpydoc_show_class_members = False
automodapi_inheritance_diagram = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_logo = "_static/btrack_logo.png"
html_theme_options = {
    "logo_only": True,
}



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import numpy.typing as npt
    from magicgui.widgets import Container

    from btrack.config import TrackerConfig
    from btrack.napari.config import TrackerConfigs

import logging

import magicgui.widgets
import napari
import qtpy.QtWidgets

import btrack
import btrack.napari.config
import btrack.napari.sync
import btrack.napari.widgets
from btrack.utils import segmentation_to_objects

__all__ = [
    "create_btrack_widget",
]

# get the logger instance
logger = logging.getLogger(__name__)

# if we don't have any handlers, set one up
if not logger.handlers:
    # configure stream handler
    log_fmt = logging.Formatter(
        "[%(levelname)s][%(asctime)s] %(message)s",
        datefmt="%Y/%m/%d %I:%M:%S %p",
    )
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(log_fmt)

    logger.addHandler(console_handler)
    logger.setLevel(logging.DEBUG)


def create_btrack_widget() -> Container:
    """Create widgets for the btrack plugin."""

    # First create our UI along with some default configs for the widgets
    all_configs = btrack.napari.config.create_default_configs()
    widgets = btrack.napari.widgets.create_widgets()
    btrack_widget = magicgui.widgets.Container(
        widgets=widgets, scrollable=True
    )
    btrack_widget.viewer = napari.current_viewer()

    # Set the cell_config defaults in the gui
    btrack.napari.sync.update_widgets_from_config(
        unscaled_config=all_configs["cell"],
        container=btrack_widget,
    )

    # Now set the callbacks
    btrack_widget.config.changed.connect(
        lambda selected: select_config(btrack_widget, all_configs, selected),
    )

    btrack_widget.call_button.changed.connect(
        lambda: run(btrack_widget, all_configs),
    )

    btrack_widget.reset_button.changed.connect(
        lambda: restore_defaults(btrack_widget, all_configs),
    )

    btrack_widget.save_config_button.changed.connect(
        lambda: save_config_to_json(btrack_widget, all_configs)
    )

    btrack_widget.load_config_button.changed.connect(
        lambda: load_config_from_json(btrack_widget, all_configs)
    )

    # there are lots of widgets so make the container scrollable
    scroll = qtpy.QtWidgets.QScrollArea()
    scroll.setWidget(btrack_widget._widget._qwidget)
    btrack_widget._widget._qwidget = scroll

    return btrack_widget


def select_config(
    btrack_widget: Container,
    configs: TrackerConfigs,
    new_config_name: str,
) -> None:
    """Set widget values from a newly-selected base config"""

    # first update the previous config with the current widget values
    previous_config_name = configs.current_config
    previous_config = configs[previous_config_name]
    previous_config = btrack.napari.sync.update_config_from_widgets(
        unscaled_config=previous_config,
        container=btrack_widget,
    )

    # now load the newly-selected config and set widget values
    configs.current_config = new_config_name
    new_config = configs[new_config_name]
    new_config = btrack.napari.sync.update_widgets_from_config(
        unscaled_config=new_config,
        container=btrack_widget,
    )


def run(btrack_widget: Container, configs: TrackerConfigs) -> None:
    """
    Update the TrackerConfig from widget values, run tracking,
    and add tracks to the viewer.
    """

    unscaled_config = configs[btrack_widget.config.current_choice]
    unscaled_config = btrack.napari.sync.update_config_from_widgets(
        unscaled_config=unscaled_config,
        container=btrack_widget,
    )

    config = unscaled_config.scale_config()
    segmentation = btrack_widget.segmentation.value
    data, properties, graph = _run_tracker(segmentation, config)

    btrack_widget.viewer.add_tracks(
        data=data,
        properties=properties,
        graph=graph,
        name=f"{segmentation}_btrack",
    )


def _run_tracker(
    segmentation: napari.layers.Image | napari.layers.Labels,
    tracker_config: TrackerConfig,
) -> tuple[npt.NDArray, dict, dict]:
    """
    Runs BayesianTracker with given segmentation and configuration.
    """
    with btrack.BayesianTracker() as tracker:
        tracker.configure(tracker_config)

        # append the objects to be tracked
        segmented_objects = segmentation_to_objects(segmentation.data)
        tracker.append(segmented_objects)

        # set the volume
        # btrack order of dimensions is XY(Z)
        # napari order of dimensions is T(Z)XY
        # so we ignore the first dimension (time) and reverse the others
        dimensions = segmentation.level_shapes[0, 1:]
        tracker.volume = tuple(
            (0, dimension) for dimension in reversed(dimensions)
        )

        # track them (in interactive mode)
        tracker.track_interactive(step_size=100)

        # generate hypotheses and run the global optimizer
        tracker.optimize()

        # get the tracks in a format for napari visualization
        data, properties, graph = tracker.to_napari(ndim=2)
        return data, properties, graph


def restore_defaults(
    btrack_widget: Container, configs: TrackerConfigs
) -> None:
    """Reload the config file then set widgets to the config's default values."""

    config_name = configs.current_config
    filename = configs[config_name].filename
    configs.add_config(
        filename=filename,
        overwrite=True,
        name=config_name,
    )

    config = configs[config_name]
    config = btrack.napari.sync.update_widgets_from_config(
        unscaled_config=config,
        container=btrack_widget,
    )


def save_config_to_json(
    btrack_widget: Container, configs: TrackerConfigs
) -> None:
    """Save widget values to file"""

    save_path = btrack.napari.widgets.save_path_dialogue_box()
    if save_path is None:
        _msg = (
            "btrack napari plugin: Configuration not saved - "
            "operation cancelled by the user."
        )
        logger.info(_msg)
        return

    unscaled_config = configs[btrack_widget.config.current_choice]
    btrack.napari.sync.update_config_from_widgets(
        unscaled_config=unscaled_config,
        container=btrack_widget,
    )
    config = unscaled_config.scale_config()

    btrack.config.save_config(save_path, config)


def load_config_from_json(
    btrack_widget: Container, configs: TrackerConfigs
) -> None:
    """Load a config from file and set it as the selected base config"""

    load_path = btrack.napari.widgets.load_path_dialogue_box()
    if load_path is None:
        _msg = "btrack napari plugin: No file loaded - operation cancelled by the user."
        logger.info(_msg)
        return

    config_name = configs.add_config(filename=load_path, overwrite=False)
    btrack_widget.config.options["choices"].append(config_name)
    btrack_widget.config.reset_choices()
    btrack_widget.config.value = config_name



import os
from typing import List, Union

import numpy as np
import pytest

import btrack

from ._utils import (
    RANDOM_SEED,
    TEST_DATA_PATH,
    create_test_object,
    create_test_segmentation_and_tracks,
)


@pytest.fixture
def test_objects():
    """
    Create a list of 10 test objects.
    """
    n_rows = 10
    return [create_test_object(test_id=i)[0] for i in range(n_rows)]


@pytest.fixture
def test_real_objects():
    """
    Create a list of objects from real data.
    """
    return btrack.io.import_CSV(TEST_DATA_PATH / "test_data.csv")


def write_h5_file(file_path: os.PathLike, test_objects) -> os.PathLike:
    """
    Write a h5 file with test objects and return path.
    """
    with btrack.io.HDF5FileHandler(file_path, "w") as h:
        h.write_objects(test_objects)

    return file_path


@pytest.fixture
def hdf5_file_path(tmp_path, test_objects) -> os.PathLike:
    """
    Create and save a btrack HDF5 file, and return the path.

    Note that this only saves segmentation results, not tracking results.
    """
    return write_h5_file(tmp_path / "test.h5", test_objects)


@pytest.fixture(params=["single", "list"])
def hdf5_file_path_or_paths(
    tmp_path, test_objects, request
) -> Union[os.PathLike, List[os.PathLike]]:
    """
    Create and save a btrack HDF5 file, and return the path.

    Note that this only saves segmentation results, not tracking results.
    """
    if request.param == "single":
        return write_h5_file(tmp_path / "test.h5", test_objects)
    elif request.param == "list":
        return [
            write_h5_file(tmp_path / "test1.h5", test_objects),
            write_h5_file(tmp_path / "test2.h5", test_objects),
        ]
    else:
        raise ValueError(
            "Invalid requests.param, must be one of 'single' or 'list'"
        )


@pytest.fixture
def test_segmentation_and_tracks():
    """
    Create a test segmentation, ground truth and example tracks.
    """
    return create_test_segmentation_and_tracks(ndim=2)


@pytest.fixture
def default_rng():
    """
    Create a default PRNG to use for tests.
    """
    return np.random.default_rng(seed=RANDOM_SEED)



import logging

from ._version import version
from .core import BayesianTracker

__all__ = ["BayesianTracker"]
__version__ = version

# get the logger instance
logger = logging.getLogger(__name__)

# if we don't have any handlers, set one up
if not logger.handlers:
    # configure stream handler
    log_fmt = logging.Formatter(
        "[%(levelname)s][%(asctime)s] %(message)s",
        datefmt="%Y/%m/%d %I:%M:%S %p",
    )
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(log_fmt)

    logger.addHandler(console_handler)
    logger.setLevel(logging.DEBUG)



#!/usr/bin/env python
# -------------------------------------------------------------------------------
# Name:     btrack
# Purpose:  A multi object tracking library, specifically used to reconstruct
#           tracks in crowded fields. Here we use a probabilistic network of
#           information to perform the trajectory linking. This method uses
#           positional and visual information for track linking.
#
# Authors:  Alan R. Lowe (arl) a.lowe@ucl.ac.uk
#
# License:  See LICENSE.md
#
# Created:  14/08/2014
# -------------------------------------------------------------------------------

__author__ = "Alan R. Lowe"
__email__ = "a.lowe@ucl.ac.uk"

import logging
from typing import List

from cvxopt import matrix, spmatrix
from cvxopt.glpk import ilp

from btrack.constants import GLPK_OPTIONS, Fates

from . import hypothesis

# get the logger instance
logger = logging.getLogger(__name__)

INIT_FATES = (
    Fates.INITIALIZE,
    Fates.INITIALIZE_BORDER,
    Fates.INITIALIZE_FRONT,
    Fates.INITIALIZE_LAZY,
)
TERM_FATES = (
    Fates.TERMINATE,
    Fates.TERMINATE_BORDER,
    Fates.TERMINATE_BACK,
    Fates.TERMINATE_LAZY,
)


class TrackOptimiser:
    """The optimizer takes a list of tracklets, as producing by a tracking
    algorithm, such as BayesianTracker, and tries to resolve small linkage
    errors, finding a globally optimal solution.

    Parameters
    ----------
    options : dict
        A dictionary of options to pass to the GLPK optimizer.


    Attributes
    ----------
    hypotheses : list[hypothesis.Hypothesis]
        A list of hypotheses generated by the hypothesis engine.

    Notes
    -----
    General types of linkage error to resolve:
        Track lost at periphery -> Terminate the track
        Track lost briefly, same state -> Link tracklets together by merging
        Object splitting (eg cell division) -> Do not merge, but create graph

    The algorithm proceeds as follows:
        1. Calculate the hypotheses for each tracklet
        2. Solve the association problem/globally optimise
        3. Merge/'link' trajectories
        4. Assign a 'fate' or optimal hypothesis to each trajectory

    The global optimisation utilises integer optimisation from the GLPK library
    Takes in the list of hypotheses and formulates as a mixed integer linear
    programming problem.  Then attempts to use GLPK to solve the association
    problem, returns a list of hypotheses to act upon.

    We set up a constraints matrix, A in this manner: (num_hypotheses x 2N)
    Uses cvxopt.gplk.ilp to:

    minimize    c'*x
    subject to  G*x <= h
                A*x = b
                x[I] are all integer (empty set in this case)
                x[B] are all binary

    Since this is a *minimisation* we need to invert the (log) probability of
    each hypothesis. If the hypotheses are badly formed, this can take *FOREVER*.

    'Report Automated Cell Lineage Construction' Al-Kofahi et al.
    Cell Cycle 2006 vol. 5 (3) pp. 327-335

    'Reliable cell tracking by global data association', Bise et al.
    2011 IEEE Symposium on Biomedical Imaging pp. 1004-1010

    'Local cellular neighbourhood controls proliferation in cell
    competition', Bove A, Gradeci D, Fujita Y, Banerjee S, Charras G and
    Lowe AR 2017 Mol. Biol. Cell vol 28 pp. 3215-3228
    """

    def __init__(self, options: dict = GLPK_OPTIONS):
        self._hypotheses: List[hypothesis.Hypothesis] = []
        self.options = options  # TODO(arl): do some option parsing?

    @property
    def hypotheses(self):
        return self._hypotheses

    @hypotheses.setter
    def hypotheses(self, hypotheses):
        self._hypotheses = hypotheses

    def optimise(self):  # noqa: PLR0915
        """Run the opimization algorithm.

        Returns
        -------
        optimized : list
            The list of selected hypotheses that forms the optimal solution.
        """

        logger.info("Setting up constraints matrix for global optimisation...")
        if self.options:
            logger.info(f"Using GLPK options: {self.options}...")

        # anon function to renumber track ID from C++
        def trk_idx(_h):
            return int(_h) - 1

        # calculate the number of hypotheses, could use this moment to cull?
        n_hypotheses = len(self.hypotheses)
        N = max(int(h.ID) for h in self.hypotheses)

        # A is the constraints matrix (store as sparse since mostly empty)
        # note that we make this in the already transposed form...
        A = spmatrix([], [], [], (2 * N, n_hypotheses), "d")
        rho = matrix(0.0, (n_hypotheses, 1), "d")

        # iterate over the hypotheses and build the constraints
        # TODO(arl): vectorize this for increased performance
        for counter, h in enumerate(self.hypotheses):
            # set the hypothesis score
            rho[counter] = h.log_likelihood

            if h.hypothesis_type == Fates.FALSE_POSITIVE:
                # is this a false positive?
                trk = trk_idx(h.ID)
                A[trk, counter] = 1
                A[N + trk, counter] = 1
                continue

            elif h.hypothesis_type in INIT_FATES:
                # an initialisation, therefore we only present this in the
                # second half of the A matrix
                trk = trk_idx(h.ID)
                A[N + trk, counter] = 1
                continue

            elif h.hypothesis_type in TERM_FATES:  # noqa: SIM114
                # a termination event, entry in first half only
                trk = trk_idx(h.ID)
                A[trk, counter] = 1
                continue

            elif h.hypothesis_type == Fates.APOPTOSIS:
                # an apoptosis event, entry in first half only
                trk = trk_idx(h.ID)
                A[trk, counter] = 1
                # A[N+trk,counter] = 1    # NOTE(arl): added 2019/08/29
                continue

            elif h.hypothesis_type == Fates.LINK:
                # a linkage event
                trk_i = trk_idx(h.ID)
                trk_j = trk_idx(h.link_ID)
                A[trk_i, counter] = 1
                A[N + trk_j, counter] = 1
                continue

            elif h.hypothesis_type == Fates.DIVIDE:
                # a branch event
                trk = trk_idx(h.ID)
                child_one = trk_idx(h.child_one_ID)
                child_two = trk_idx(h.child_two_ID)
                A[trk, counter] = 1
                A[N + child_one, counter] = 1
                A[N + child_two, counter] = 1
                continue

            elif h.hypothesis_type == Fates.MERGE:
                # a merge event
                trk = trk_idx(h.ID)
                parent_one = trk_idx(h.parent_one_ID)
                parent_two = trk_idx(h.parent_two_ID)
                A[N + trk, counter] = 1
                A[parent_one, counter] = 1
                A[parent_two, counter] = 1
                continue

            else:
                raise ValueError(f"Unknown hypothesis: {h.hypothesis_type}")

        logger.info("Optimizing...")

        # now set up the ILP solver
        G = spmatrix([], [], [], (2 * N, n_hypotheses), "d")
        h = matrix(0.0, (2 * N, 1), "d")  # NOTE h cannot be a sparse matrix
        Ix = set()  # empty set of x which are integer
        B = set(range(n_hypotheses))  # signifies all are binary in x
        b = matrix(1.0, (2 * N, 1), "d")

        # now try to solve it!!!
        status, x = ilp(-rho, -G, h, A, b, Ix, B, options=self.options)

        # log the warning if not optimal solution
        if status != "optimal":
            logger.warning(f"Optimizer returned status: {status}")
            return []

        # return only the selected hypotheses
        results = [i for i, h in enumerate(self.hypotheses) if x[i] > 0]

        logger.info(f"Optimization complete. (Solution: {status})")
        return results



"""
This module is a reader plugin btrack files for napari.
"""
import os
from typing import Callable, List, Optional, Sequence, Union

from napari.types import LayerDataTuple
from napari_plugin_engine import napari_hook_implementation

from btrack.io import HDF5FileHandler
from btrack.utils import tracks_to_napari

# Type definitions
PathOrPaths = Union[os.PathLike, Sequence[os.PathLike]]
ReaderFunction = Callable[[PathOrPaths], List[LayerDataTuple]]


@napari_hook_implementation
def get_reader(path: PathOrPaths) -> Optional[ReaderFunction]:
    """A basic implementation of the napari_get_reader hook specification.

    Parameters
    ----------
    path : str or list of str
        Path to file, or list of paths.

    Returns
    -------
    function or None
        If the path is a recognized format, return a function that accepts the
        same path or list of paths, and returns a list of layer data tuples.
    """
    return reader_function


def reader_function(path: PathOrPaths) -> List[LayerDataTuple]:
    """Take a path or list of paths and return a list of LayerData tuples.

    Readers are expected to return data as a list of tuples, where each tuple
    is (data, [add_kwargs, [layer_type]]), "add_kwargs" and "layer_type" are
    both optional.

    Parameters
    ----------
    path : str or list of str
        Path to file, or list of paths.

    Returns
    -------
    layer_data : list of tuples
        A list of LayerData tuples where each tuple in the list contains
        (data, metadata, layer_type), where data is a numpy array, metadata is
        a dict of keyword arguments for the corresponding viewer.add_* method
        in napari, and layer_type is a lower-case string naming the type of
        layer. Both "metadata" and "layer_type" are optional. napari will default
        to layer_type=="image" if not provided
    """
    # handle both a string and a list of strings
    paths = [path] if not isinstance(path, list) else path

    # store the layers to be generated
    layers: List[tuple] = []

    for _path in paths:
        with HDF5FileHandler(_path, "r") as hdf:
            # get the segmentation if there is one
            segmentation = hdf.segmentation
            if segmentation is not None:
                layers.append((segmentation, {}, "labels"))

            # iterate over object types and create a layer for each
            for obj_type in hdf.object_types:
                # set the object type, and retrieve the tracks
                hdf.object_type = obj_type

                if f"tracks/{obj_type}" not in hdf._hdf:
                    continue

                tracklets = hdf.tracks
                tracks, properties, graph = tracks_to_napari(tracklets)

                # optional kwargs for the corresponding viewer.add_* method
                # https://napari.org/docs/api/napari.components.html#module-napari.components.add_layers_mixin
                add_kwargs = {
                    "properties": properties,
                    "graph": graph,
                    "name": obj_type,
                    "blending": "translucent",
                }

                layer = (tracks, add_kwargs, "tracks")
                layers.append(layer)
    return layers



#!/usr/bin/env python
# -------------------------------------------------------------------------------
# Name:     btrack
# Purpose:  A multi object tracking library, specifically used to reconstruct
#           tracks in crowded fields. Here we use a probabilistic network of
#           information to perform the trajectory linking. This method uses
#           positional and visual information for track linking.
#
# Authors:  Alan R. Lowe (arl) a.lowe@ucl.ac.uk
#
# License:  See LICENSE.md
#
# Created:  14/08/2014
# -------------------------------------------------------------------------------

__author__ = "Alan R. Lowe"
__email__ = "a.lowe@ucl.ac.uk"


import ctypes

from btrack import constants

H_TYPES = [
    "P_FP",
    "P_init",
    "P_term",
    "P_link",
    "P_branch",
    "P_dead",
    "P_merge",
]


class Hypothesis(ctypes.Structure):
    """Hypothesis structure.

    This is essentially a replicate of the C-structure used by the hypothesis
    engine. Added a few extra functions to enable quick assignment.

    unsigned int hypothesis;
    unsigned int ID;
    double probability;
    unsigned int link_ID;
    unsigned int child_one_ID;
    unsigned int child_two_ID;

    Notes:
      These are automatically generated by the optimiser
    """

    _fields_ = [
        ("hypothesis", ctypes.c_uint),
        ("ID", ctypes.c_uint),
        ("probability", ctypes.c_double),
        ("link_ID", ctypes.c_uint),
        ("child_one_ID", ctypes.c_uint),
        ("child_two_ID", ctypes.c_uint),
        ("parent_one_ID", ctypes.c_uint),
        ("parent_two_ID", ctypes.c_uint),
    ]

    @property
    def hypothesis_type(self):
        return constants.Fates(self.hypothesis)

    @property
    def log_likelihood(self):
        return self.probability


class PyHypothesisParams(ctypes.Structure):
    """HypothesisParams.

    A convenient way to send hypothesis generation parameters to the hypothesis
    engine.

      double lambda_time;   // lambda_1
      double lambda_dist;   // lambda_2
      double lambda_link;   // lambda_3
      double lambda_branch; // lambda_4
      double eta;           // default low probability
      double theta_dist;
      double theta_time;
      double dist_thresh;
      double time_thresh;
      unsigned int apop_thresh;
      double segmentation_miss_rate;
      double apoptosis_rate;
      bool relax;

    Notes:
        None

    """

    _fields_ = [
        ("lambda_time", ctypes.c_double),
        ("lambda_dist", ctypes.c_double),
        ("lambda_link", ctypes.c_double),
        ("lambda_branch", ctypes.c_double),
        ("eta", ctypes.c_double),
        ("theta_dist", ctypes.c_double),
        ("theta_time", ctypes.c_double),
        ("dist_thresh", ctypes.c_double),
        ("time_thresh", ctypes.c_double),
        ("apop_thresh", ctypes.c_uint),
        ("segmentation_miss_rate", ctypes.c_double),
        ("apoptosis_rate", ctypes.c_double),
        ("relax", ctypes.c_bool),
        ("hypotheses_to_generate", ctypes.c_uint),
    ]



from __future__ import annotations

import itertools
import logging
import os
import re
from functools import wraps
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import h5py
import numpy as np

# import core
from btrack import _version, btypes, constants, utils

from .utils import (
    check_object_type,
    check_track_type,
    objects_from_array,
    objects_from_dict,
)

if TYPE_CHECKING:
    from btrack import BayesianTracker

# get the logger instance
logger = logging.getLogger(__name__)


def h5check_property_exists(property):  # noqa: A002
    """Wrapper for hdf handler to make sure a property exists."""

    def func(fn):
        @wraps(fn)
        def wrapped_handler_property(*args, **kwargs):
            self = args[0]
            assert isinstance(self, HDF5FileHandler)
            if property not in self._hdf:
                logger.error(
                    f"{property.capitalize()} not found in {self.filename}"
                )
                return None
            return fn(*args, **kwargs)

        return wrapped_handler_property

    return func


class HDF5FileHandler:
    """Generic HDF5 file hander for reading and writing datasets. This is
    inter-operable between segmentation, tracking and analysis code.

    Parameters
    ----------
    filename : str
        The filename of the hdf5 file to be used.
    read_write : str
        A read/write mode for the file, e.g. `w`, `r`, `a` etc.
    obj_type : str
        The name of the object type. Defaults to `obj_type_1`. The object type
        name must start with `obj_type_`

    Attributes
    ----------
    segmentation : np.ndarray
        A numpy array representing the segmentation data. TZYX
    objects : list [PyTrackObject]
        A list of PyTrackObjects localised from the segmentation data.
    filtered_objects  : np.ndarray
        Similar to objects, but filtered by property.
    tracks : list [Tracklet]
        A list of Tracklet objects.
    lbep : np.ndarray
        The LBEP table representing the track graph.

    Notes
    -----
    Basic format of the HDF file is::

        data.h5/
        ├─ segmentation/
        │  ├─ images                (J x (d) x h x w) uint16 segmentation
        ├─ objects/
        │  ├─ obj_type_1/
        │  │  ├─ coords             (I x 5) [t, x, y, z, object_type]
        │  │  ├─ labels             (I x D) [label, (softmax scores ...)]
        │  │  ├─ map                (J x 2) [start_index, end_index] -> coords array
        │  │  ├─ properties /
        │  │  │  ├─ area            (I x 1) first named property (e.g. `area`)
        │  │  │  ├─ ...
        │  ├─ obj_type_2/
        │  ├─ ...
        ├─ tracks/
        │  ├─ obj_type_1/
        │  │  ├─ tracks             (I x 1) [index into coords]
        │  │  ├─ dummies            similar to objects/coords, but for dummy objects
        │  │  ├─ map                (K x 2) [start_index, end_index] -> tracks array
        │  │  ├─ LBEPRG             (K x 6) [L, B, E, P, R, G]
        │  │  ├─ fates              (K x n) [fate_from_tracker, ...future_expansion]
        │  ├─ obj_type_2/
        │  ├─ ...

    Where ``I`` is the number of objects, ``J`` is the number of frames and
    ``K`` is number of tracks. LBEPR is a modification of the LBEP format to
    also include the root node of the tree.


    Examples
    --------
    Read objects from a file:

    >>> with HDF5FileHandler('file.h5', 'r') as handler:
    ...    objects = handler.objects

    Use filtering by property for object retrieval:

    >>> obj = handler.filtered_objects('flag==1')
    ... obj = handler.filtered_objects('area>100')

    Write tracks directly to a file:

    >>> handler.write_tracks(tracks)

    """

    def __init__(
        self,
        filename: os.PathLike,
        read_write: str = "r",
        *,
        obj_type: str = "obj_type_1",
    ):
        self._f_expr = None  # DO NOT USE
        self.object_type = obj_type

        self.filename = filename
        logger.info(f"Opening HDF file: {self.filename}...")
        self._hdf = h5py.File(filename, read_write)
        self._states = list(constants.States)

    @property
    def object_types(self) -> List[str]:
        return list(self._hdf["objects"].keys())

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        if not self._hdf:
            return
        logger.info(f"Closing HDF file: {self.filename}")
        self._hdf.close()

    @property
    def object_type(self) -> str:
        return self._object_type

    @object_type.setter
    def object_type(self, obj_type: str) -> None:
        if not obj_type.startswith("obj_type_"):
            raise ValueError("Object type must start with ``obj_type_``")
        self._object_type = obj_type

    @property  # type: ignore
    @h5check_property_exists("segmentation")
    def segmentation(self) -> np.ndarray:
        segmentation = self._hdf["segmentation"]["images"][:].astype(np.uint16)
        logger.info(f"Loading segmentation {segmentation.shape}")
        return segmentation

    def write_segmentation(self, segmentation: np.ndarray) -> None:
        """Write out the segmentation to an HDF file.

        Parameters
        ----------
        segmentation : np.ndarray
            A numpy array representing the segmentation data. T(Z)YX, uint16
        """
        # write the segmentation out
        grp = self._hdf.create_group("segmentation")
        grp.create_dataset(
            "images",
            data=segmentation,
            dtype="uint16",
            compression="gzip",
            compression_opts=7,
        )

    @property
    def objects(self) -> List[btypes.PyTrackObject]:
        """Return the objects in the file."""
        return self.filtered_objects()

    @h5check_property_exists("objects")
    def filtered_objects(
        self,
        f_expr: Optional[str] = None,
        *,
        lazy_load_properties: bool = True,
        exclude_properties: Optional[List[str]] = None,
    ) -> List[btypes.PyTrackObject]:
        """A filtered list of objects based on metadata.

        Parameters
        ----------
        f_expr : str
            A string representing a filtering option. For example, `area>100`
            would filter objects by a property key `area` where the numerical
            value of area was greater than 100.
        lazy_load_properties : bool
            For future expansion. To allow lazy loading of large datasets.
        exclude_properties : list or None
            A list of properties keys to exclude when loading from disk.

        Returns
        -------
        objects : list
            A list of :py:class:`btrack.btypes.PyTrackObject` objects.
        """

        exclude_properties = (
            exclude_properties if exclude_properties is not None else []
        )

        if self.object_type not in self.object_types:
            raise ValueError(f"Object type {self.object_type} not recognized")

        grp = self._hdf["objects"][self.object_type]

        # read the whole dataset into memory
        txyz = grp["coords"][:]
        if "labels" not in grp:
            logger.warning("Labels missing from objects in HDF file")
            labels = np.zeros((txyz.shape[0], 6))
        else:
            labels = self._hdf["objects"][self.object_type]["labels"][:]

        # get properties if we have them (note, this assumes that the same
        # properties exist for each object)
        properties = {}
        if "properties" in grp:
            p_keys = list(
                set(grp["properties"].keys()).difference(
                    set(exclude_properties)
                )
            )
            properties = {k: grp["properties"][k][:] for k in p_keys}
            assert all([len(p) == len(txyz) for p in properties.values()])

        # note that this doesn't do much error checking at the moment
        # TODO(arl): this should now reference the `properties`
        if f_expr is not None:
            assert isinstance(f_expr, str)
            pattern = r"(?P<name>\w+)(?P<op>[\>\<\=]+)(?P<cmp>[0-9]+)"
            m = re.match(pattern, f_expr)

            if m is None:
                raise ValueError(f"Cannot filter objects by {f_expr}")

            f_eval = f"x{m['op']}{m['cmp']}"  # e.g. x > 10

            if m["name"] in properties:
                data = properties[m["name"]]
                filtered_idx = [i for i, x in enumerate(data) if eval(f_eval)]
            else:
                raise ValueError(f"Cannot filter objects by {f_expr}")

        else:
            filtered_idx = range(txyz.shape[0])  # default filtering uses all

        # sanity check that coordinates matches labels
        assert txyz.shape[0] == labels.shape[0]
        logger.info(
            f"Loading objects/{self.object_type} {txyz.shape} "
            f"({len(filtered_idx)} filtered: {f_expr})"
        )

        txyz_filtered = txyz[filtered_idx, :]
        labels_filtered = labels[filtered_idx, :]

        objects_dict = {
            "t": txyz_filtered[:, 0],
            "x": txyz_filtered[:, 1],
            "y": txyz_filtered[:, 2],
            "z": txyz_filtered[:, 3],
            "label": labels_filtered[:, 0],
            "ID": np.asarray(filtered_idx),
        }

        # add the filtered properties
        for key, props in properties.items():
            objects_dict.update({key: props[filtered_idx]})

        return objects_from_dict(objects_dict)

    def write_objects(
        self, data: Union[List[btypes.PyTrackObject], BayesianTracker]
    ) -> None:
        """Write objects to HDF file.

        Parameters
        ----------
        data : list or BayesianTracker instance
            Either a list of PyTrackObject to be written, or an instance of
            BayesianTracker with a .objects property.
        """
        # TODO(arl): make sure that the objects are ordered in time

        if isinstance(data, list):
            objects = data
        elif hasattr(data, "objects"):
            objects = data.objects
        else:
            raise TypeError("Object type not recognized.")

        # make sure that the data to be written are all of type PyTrackObject
        if not check_object_type(objects):
            raise TypeError("Object type not recognized.")

        if "objects" not in self._hdf:
            self._hdf.create_group("objects")
        grp = self._hdf["objects"].create_group(self.object_type)
        props = {k: [] for k in objects[0].properties}

        n_objects = len(objects)
        n_frames = np.max([o.t for o in objects]) + 1

        txyz = np.zeros((n_objects, 5), dtype=np.float32)
        labels = np.zeros((n_objects, 1), dtype=np.uint8)
        fmap = np.zeros((n_frames, 2), dtype=np.uint32)

        # convert the btrack objects into a numpy array
        for i, obj in enumerate(objects):
            txyz[i, :] = [obj.t, obj.x, obj.y, obj.z, 0]
            labels[i, :] = obj.label
            t = int(obj.t)
            fmap[t, 1] = np.max([fmap[t, 1], i])

            # add in any properties
            for key in props:
                props[key].append(obj.properties[key])

        fmap[1:, 0] = fmap[:-1, 1]

        logger.info(f"Writing objects/{self.object_type}")
        grp.create_dataset("coords", data=txyz, dtype="float32")
        grp.create_dataset("map", data=fmap, dtype="uint32")

        logger.info(f"Writing labels/{self.object_type}")
        grp.create_dataset("labels", data=labels, dtype="float32")

        # finally, write any properties
        self.write_properties(props)

    @h5check_property_exists("objects")
    def write_properties(
        self, data: Dict[str, Any], *, allow_overwrite: bool = False
    ) -> None:
        """Write object properties to HDF file.

        Parameters
        ----------
        data : dict {key: (N, D)}
            A dictionary of key-value pairs of properties to be written. The
            values should be an array equal in length to the number of objects
            and with D dimensions.
        allow_overwrite : bool
            Allow to delete the existing property keys from the HDF5 file and
            overwrite with new values from the data dict. Defaults to False.
        """

        if not isinstance(data, dict):
            raise TypeError("Properties must be a dictionary.")

        grp = self._hdf[f"objects/{self.object_type}"]

        if "properties" not in grp.keys():
            props_grp = grp.create_group("properties")
        else:
            props_grp = self._hdf[f"objects/{self.object_type}/properties"]

        n_objects = len(self.objects)

        for key, values in data.items():
            # Manage the property data:
            if not values:
                logger.warning(f"Property '{key}' is empty.")
                continue
            values = np.array(values)  # noqa: PLW2901
            assert values.shape[0] == n_objects

            # Check if the property is already in the props_grp:
            if key in props_grp:
                if allow_overwrite is False:
                    logger.info(
                        f"Property '{key}' already written in the file"
                    )
                    raise KeyError(
                        f"Property '{key}' already in file -> switch on "
                        "'overwrite' param to replace existing property "
                    )
                else:
                    del self._hdf[f"objects/{self.object_type}/properties"][
                        key
                    ]
                    logger.info(
                        f"Property '{key}' erased to be overwritten..."
                    )

            # Now that you handled overwriting, write the values:
            logger.info(
                f"Writing properties/{self.object_type}/{key} {values.shape}"
            )
            props_grp.create_dataset(key, data=data[key], dtype="float32")

    @property  # type: ignore
    @h5check_property_exists("tracks")
    def tracks(self) -> List[btypes.Tracklet]:
        """Return the tracks in the file."""

        logger.info(f"Loading tracks/{self.object_type}")
        track_map = self._hdf["tracks"][self.object_type]["map"][:]
        track_refs = self._hdf["tracks"][self.object_type]["tracks"][:]
        lbep = self.lbep
        fates = self._hdf["tracks"][self.object_type]["fates"][:]

        # if there are dummies, make new dummy objects
        if "dummies" in self._hdf["tracks"][self.object_type]:
            dummies = self._hdf["tracks"][self.object_type]["dummies"][:]
            dummy_obj = objects_from_array(dummies[:, :4])
            for d in dummy_obj:
                d.ID = -(d.ID + 1)  # restore the -ve ID
                d.dummy = True  # set the dummy flag to true

        # TODO(arl): this needs to be stored in the HDF folder
        if (
            "f_expr" in self._hdf["tracks"][self.object_type].attrs
            and self._f_expr is None
        ):
            f_expr = self._hdf["tracks"][self.object_type].attrs["f_expr"]
        elif self._f_expr is not None:
            f_expr = self._f_expr
        else:
            f_expr = None

        obj = self.filtered_objects(f_expr=f_expr)

        def _get_txyz(_ref: int) -> int:
            if _ref >= 0:
                return obj[_ref]
            return dummy_obj[abs(_ref) - 1]  # references are -ve for dummies

        tracks = []
        for i in range(track_map.shape[0]):
            idx = slice(*track_map[i, :].tolist())
            refs = track_refs[idx]
            track = btypes.Tracklet(lbep[i, 0], list(map(_get_txyz, refs)))
            track.parent = lbep[i, 3]  # set the parent and root of tree
            track.root = lbep[i, 4]
            if lbep.shape[1] > 5:  # noqa: PLR2004
                track.generation = lbep[i, 5]
            track.fate = constants.Fates(fates[i])  # restore the track fate
            tracks.append(track)

        # once we have all of the tracks, populate the children
        to_update = {}
        for track in tracks:
            if not track.is_root:
                parents = filter(lambda t: track.parent == t.ID, tracks)
                for parent in parents:
                    if parent not in to_update:
                        to_update[parent] = []
                    to_update[parent].append(track.ID)

        # sanity check, can be removed at a later date
        MAX_N_CHILDREN = 2
        assert all(
            len(children) <= MAX_N_CHILDREN for children in to_update.values()
        )

        # add the children to the parent
        for track, children in to_update.items():
            track.children = children

        return tracks

    def write_tracks(  # noqa: PLR0912
        self,
        data: Union[List[btypes.Tracklet], BayesianTracker],
        *,
        f_expr: Optional[str] = None,
    ) -> None:
        """Write tracks to HDF file.

        Parameters
        ----------
        data : list of Tracklets or an instance of BayesianTracker
            A list of tracklets or an instance of BayesianTracker.
        f_expr : str
            An expression which represents how the objects have been filtered
            prior to tracking, e.g. `area>100.0`
        """

        if isinstance(data, list):
            if not check_track_type(data):
                raise ValueError(f"Data of type {type(data)} not supported.")

            all_objects = itertools.chain.from_iterable(
                [trk._data for trk in data]
            )

            objects = [obj for obj in all_objects if not obj.dummy]
            dummies = [obj for obj in all_objects if obj.dummy]

            # renumber the object ID so that they can be stored in a contiguous
            # array and indexed by row - this may not be necessary for most
            # datasets, but is here just in case
            for idx, obj in enumerate(objects):
                obj.ID = idx

            for idx, dummy in enumerate(dummies):
                dummy.ID = -(idx + 1)

            refs = [trk.refs for trk in data]
            lbep_table = utils._lbep_table(data)
            fate_table = np.stack([t.fate.value for t in data], axis=0)

            if "objects" not in self._hdf:
                self.write_objects(objects)

        elif hasattr(data, "tracks"):
            refs = data.refs
            dummies = data.dummies
            lbep_table = data.LBEP
            fate_table = np.stack([t.fate.value for t in data.tracks], axis=0)
        else:
            raise ValueError(f"Data of type {type(data)} not supported.")

        if not refs:
            logger.error(f"No tracks found when exporting to: {self.filename}")
            return

        # sanity check
        assert lbep_table.shape[0] == len(refs)

        logger.info(f"Writing tracks/{self.object_type}")
        hdf_tracks = np.concatenate(refs, axis=0)

        hdf_frame_map = np.zeros((len(refs), 2), dtype=np.int32)
        for i, track in enumerate(refs):
            offset = hdf_frame_map[i - 1, 1] if i > 0 else 0
            hdf_frame_map[i, :] = np.array([0, len(track)]) + offset

        if "tracks" not in self._hdf:
            self._hdf.create_group("tracks")

        if self.object_type in self._hdf["tracks"]:
            logger.warning(f"Removing tracks/{self.object_type}.")
            del self._hdf["tracks"][self.object_type]

        grp = self._hdf["tracks"].create_group(self.object_type)
        grp.create_dataset("tracks", data=hdf_tracks, dtype="int32")
        grp.create_dataset("map", data=hdf_frame_map, dtype="uint32")

        # if we have used the f_expr we can save it as an attribute here
        if f_expr is not None and isinstance(f_expr, str):
            grp.attrs["f_expr"] = f_expr

        # also save the version number as an attribute
        grp.attrs["version"] = _version.version

        # write out dummies
        if dummies:
            logger.info(f"Writing dummies/{self.object_type}")
            o = self.object_types.index(self.object_type) + 1
            txyz = np.stack([[d.t, d.x, d.y, d.z, o] for d in dummies], axis=0)
            grp.create_dataset("dummies", data=txyz, dtype="float32")

        # write out the LBEP table
        logger.info(f"Writing LBEP/{self.object_type}")
        grp.create_dataset("LBEPR", data=lbep_table, dtype="int32")

        # write out cell fates
        logger.info(f"Writing fates/{self.object_type}")
        grp.create_dataset("fates", data=fate_table, dtype="int32")

    @property  # type: ignore
    @h5check_property_exists("tracks")
    def lbep(self) -> np.ndarray:
        """Return the LBEP data."""
        logger.info(f"Loading LBEP/{self.object_type}")
        return self._hdf["tracks"][self.object_type]["LBEPR"][:]



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from magicgui.widgets import Widget

from btrack.napari.widgets._general import (
    create_control_widgets,
    create_input_widgets,
    create_update_method_widgets,
)
from btrack.napari.widgets._hypothesis import create_hypothesis_model_widgets
from btrack.napari.widgets._motion import create_motion_model_widgets


def create_widgets() -> list[Widget]:
    """Create all the widgets for the plugin"""

    input_widgets = create_input_widgets()
    update_method_widgets = create_update_method_widgets()
    motion_model_widgets = create_motion_model_widgets()
    hypothesis_model_widgets = create_hypothesis_model_widgets()
    control_buttons = create_control_widgets()

    return [
        *input_widgets,
        *update_method_widgets,
        *motion_model_widgets,
        *hypothesis_model_widgets,
        *control_buttons,
    ]



from pathlib import Path
from typing import Tuple, Union

import numpy as np
import pytest
from pydantic import BaseModel

import btrack

from ._utils import CONFIG_FILE, RANDOM_SEED


def _random_config() -> dict:
    rng = np.random.default_rng(seed=RANDOM_SEED)
    return {
        "max_search_radius": rng.uniform(1, 100),
        "update_method": rng.choice(btrack.constants.BayesianUpdates),
        "return_kalman": bool(rng.uniform(0, 2)),
        "store_candidate_graph": bool(rng.uniform(0, 2)),
        "verbose": bool(rng.uniform(0, 2)),
        "volume": tuple([(0, rng.uniform(1, 100)) for _ in range(3)]),
    }


def _validate_config(
    cfg: Union[btrack.BayesianTracker, BaseModel], options: dict
):
    for key, value in options.items():
        cfg_value = getattr(cfg, key)
        # takes care of recursive model definintions (i.e. MotionModel inside
        # TrackerConfig).
        if isinstance(cfg_value, BaseModel):
            _validate_config(cfg_value, value)
        else:
            np.testing.assert_equal(cfg_value, value)


def test_config():
    """Test creation of a TrackerConfig object."""
    options = _random_config()
    cfg = btrack.config.TrackerConfig(**options)
    assert isinstance(cfg, btrack.config.TrackerConfig)

    for key, value in options.items():
        np.testing.assert_equal(getattr(cfg, key), value)


def test_export_config(tmp_path):
    """Test exporting the config."""
    options = _random_config()
    cfg = btrack.config.TrackerConfig(**options)

    # export it
    tmp_cfg_file = Path(tmp_path) / "config.json"
    btrack.config.save_config(tmp_cfg_file, cfg)
    assert tmp_cfg_file.exists()


def test_import_config():
    """Test loading a config from a file."""
    cfg = btrack.config.load_config(CONFIG_FILE)
    assert isinstance(cfg, btrack.config.TrackerConfig)


def test_config_to_json():
    """Test that a config can be converted to json format without raising an error"""
    cfg = btrack.config.load_config(CONFIG_FILE)
    cfg.json()


def test_config_tracker_setters():
    """Test configuring the tracker using setters."""
    options = _random_config()
    with btrack.BayesianTracker() as tracker:
        # use the setters to apply the comfiguration
        for key, value in options.items():
            setattr(tracker, key, value)

        # use the getters or the config
        _validate_config(tracker, options)
        _validate_config(tracker.configuration, options)


def _cfg_dict() -> Tuple[dict, dict]:
    cfg_raw = btrack.config.load_config(CONFIG_FILE)
    cfg = _random_config()
    cfg.update(cfg_raw.dict())
    assert isinstance(cfg, dict)
    return cfg, cfg


def _cfg_file() -> Tuple[Path, dict]:
    filename = CONFIG_FILE
    assert isinstance(filename, Path)
    cfg = btrack.config.load_config(filename)
    return filename, cfg.dict()


def _cfg_pydantic() -> Tuple[btrack.config.TrackerConfig, dict]:
    cfg = btrack.config.load_config(CONFIG_FILE)
    options = _random_config()
    for key, value in options.items():
        setattr(cfg, key, value)
    assert isinstance(cfg, btrack.config.TrackerConfig)
    return cfg, cfg.dict()


@pytest.mark.parametrize("get_cfg", [_cfg_file, _cfg_dict, _cfg_pydantic])
def test_config_tracker(get_cfg):
    """Test configuring the tracker from a file, dictionary or `TrackerConfig`."""
    cfg, cfg_dict = get_cfg()

    with btrack.BayesianTracker() as tracker:
        tracker.configure(cfg)
        _validate_config(tracker, cfg_dict)



import numpy as np
import pytest

from ._utils import (
    create_test_object,
    create_test_properties,
    create_test_tracklet,
)


def test_create_object():
    """Test that an object is correctly instantiated, and that the stored
    data matches the data used for creation."""
    obj, data = create_test_object()
    for k, v in data.items():
        np.testing.assert_equal(getattr(obj, k), v)


def test_object_to_dict():
    """Test that an object can be converted to a dictionary using the built-in
    method."""
    obj, data = create_test_object()
    obj_as_dict = obj.to_dict()
    for k, v in data.items():
        np.testing.assert_equal(obj_as_dict[k], v)


@pytest.mark.parametrize("properties", [{}, create_test_properties()])
def test_create_object_with_properties(properties: dict):
    """Test an object with some properties."""
    obj, data = create_test_object()
    obj.properties = properties
    for k, v in properties.items():
        np.testing.assert_equal(obj.properties[k], v)


@pytest.mark.parametrize("properties", [{}, create_test_properties()])
def test_object_features(properties: dict):
    """Test creating object and setting tracking features."""
    obj, _ = create_test_object()
    obj.properties = properties
    assert obj.n_features == 0
    keys = list(properties.keys())
    obj.set_features(keys)
    n_keys = sum([np.asarray(p).size for p in properties.values()])
    assert obj.n_features == n_keys


def test_object_feature_values():
    """Test creating object with features and returning correct values."""
    obj, _ = create_test_object()
    properties = create_test_properties()
    obj.properties = properties
    keys = list(properties.keys())
    obj.set_features(keys)

    # check that the numpy data matches
    raw_features = np.ctypeslib.as_array(obj.features, shape=(obj.n_features,))
    flat_properties = np.concatenate(
        [np.asarray(obj.properties[k]).ravel() for k in keys], axis=0
    )
    np.testing.assert_almost_equal(flat_properties, raw_features)


def test_object_incorrect_features():
    """Test creating object and setting tracking features."""
    obj, data = create_test_object()
    assert obj.n_features == 0
    with pytest.raises(KeyError):
        obj.set_features(
            [
                "this_key_does_not_exist",
            ]
        )
    assert obj.n_features == 0


@pytest.mark.parametrize("track_len", [0, 1, 10, 100, 1000])
def test_create_tracklet(track_len: int):
    """Test that a track is correctly instantiated, and that the stored
    data matches the data used for creation."""
    tracklet, data, properties, track_ID = create_test_tracklet(track_len)
    assert len(tracklet) == len(data)

    # now check that the track data is correct
    assert track_ID == tracklet.ID
    fields = ["x", "y", "z", "t"]
    for field in fields:
        obj_data = [getattr(obj, field) for obj in data]
        np.testing.assert_equal(obj_data, getattr(tracklet, field))


@pytest.mark.parametrize("track_len", [0, 1, 10, 100, 1000])
def test_create_tracklet_with_properties(track_len: int):
    """Test that a track is correctly instantiated, and that the stored
    properties match the data used for creation."""
    tracklet, data, properties, track_ID = create_test_tracklet(track_len)
    t_properties = tracklet.properties

    for k, v in properties.items():
        np.testing.assert_equal(t_properties[k], v)


def test_tracklet_to_dict():
    """Test that a tracklet can be converted to a dictionary using the built-in
    method."""
    track_len = 10
    tracklet, data, properties, track_ID = create_test_tracklet(track_len)
    tracklet_as_dict = tracklet.to_dict()
    # now check that the track data is correct
    assert track_ID == tracklet_as_dict["ID"]
    fields = ["x", "y", "z", "t"]
    for field in fields:
        obj_data = [getattr(obj, field) for obj in data]
        np.testing.assert_equal(obj_data, tracklet_as_dict[field])



import ctypes
import logging
import os
import platform

import numpy as np

from btrack.btypes import PyGraphEdge, PyTrackingInfo, PyTrackObject
from btrack.constants import BTRACK_PATH
from btrack.optimise import hypothesis

# get the logger instance
logger = logging.getLogger(__name__)


def numpy_pointer_decorator(func):
    """simple decorator for numpy ctypes pointers"""
    return func()


@numpy_pointer_decorator
def np_dbl_p():
    """Temporary function. Will remove in final release"""
    return np.ctypeslib.ndpointer(
        dtype=np.double, ndim=2, flags="C_CONTIGUOUS"
    )


@numpy_pointer_decorator
def np_dbl_pc():
    """Temporary function. Will remove in final release"""
    return np.ctypeslib.ndpointer(
        dtype=np.double, ndim=2, flags="F_CONTIGUOUS"
    )


@numpy_pointer_decorator
def np_uint_p():
    """Temporary function. Will remove in final release"""
    return np.ctypeslib.ndpointer(
        dtype=np.uint32, ndim=2, flags="C_CONTIGUOUS"
    )


@numpy_pointer_decorator
def np_int_p():
    """Temporary function. Will remove in final release"""
    return np.ctypeslib.ndpointer(dtype=np.int32, ndim=2, flags="C_CONTIGUOUS")


@numpy_pointer_decorator
def np_int_vec_p():
    """Temporary function. Will remove in final release"""
    return np.ctypeslib.ndpointer(
        dtype=np.int32, ndim=1
    )  # , flags='C_CONTIGUOUS')


def load_library(filename):
    """Return the platform for shared library loading.

    Parameters
    ----------
    filename : str
        Filename for the shared library.

    Returns
    -------
    lib : ctypes.CDLL
        The ctypes `btrack` library.
    """

    if not isinstance(filename, str):
        raise TypeError("Filename must be a string")

    lib_file, ext = os.path.splitext(filename)

    system = platform.system()

    file_ext = {"Linux": ".so", "Darwin": ".dylib", "Windows": ".DLL"}
    full_lib_file = lib_file + file_ext[system]

    try:
        lib = ctypes.cdll.LoadLibrary(full_lib_file)
        logger.info(f"Loaded btrack: {full_lib_file}")
    except OSError as err:
        raise OSError(f"Cannot load shared library {full_lib_file}") from err

    return lib


def get_library():  # noqa: PLR0915
    """Loads and returns the btrack shared library."""
    lib = load_library(os.path.join(BTRACK_PATH, "libs", "libtracker"))

    # deal with constructors/destructors
    lib.new_interface.restype = ctypes.c_void_p
    lib.new_interface.argtypes = [ctypes.c_bool]

    lib.del_interface.restype = None
    lib.del_interface.argtypes = [ctypes.c_void_p]

    # set the update method
    lib.set_update_mode.restype = None
    lib.set_update_mode.argtypes = [ctypes.c_void_p, ctypes.c_uint]

    # set the features to use during the update method
    lib.set_update_features.restype = None
    lib.set_update_features.argtypes = [ctypes.c_void_p, ctypes.c_uint]

    # set the tracker to store the candidate graph
    lib.set_store_candidate_graph.restype = None
    lib.set_store_candidate_graph.argtypes = [ctypes.c_void_p, ctypes.c_bool]

    # set the motion model
    lib.motion.restype = None
    lib.motion.argtypes = [
        ctypes.c_void_p,
        ctypes.c_uint,
        ctypes.c_uint,
        np_dbl_p,
        np_dbl_p,
        np_dbl_p,
        np_dbl_p,
        np_dbl_p,
        ctypes.c_double,
        ctypes.c_double,
        ctypes.c_uint,
        ctypes.c_double,
    ]

    # set the object model
    # lib.model.restype = None
    # lib.model.argtypes = [ctypes.c_void_p, ctypes.c_uint, np_dbl_p,
    #                         np_dbl_p, np_dbl_p]

    lib.max_search_radius.restype = None
    lib.max_search_radius.argtypes = [ctypes.c_void_p, ctypes.c_float]

    # append a new observation
    lib.append.restype = None
    lib.append.argtypes = [ctypes.c_void_p, PyTrackObject]

    # run the complete tracking
    lib.track.restype = ctypes.POINTER(PyTrackingInfo)
    lib.track.argtypes = [ctypes.c_void_p]

    # run one or more steps of the tracking, interactive mode
    lib.step.restype = ctypes.POINTER(PyTrackingInfo)
    lib.step.argtypes = [ctypes.c_void_p, ctypes.c_uint]

    # get an individual track length
    lib.track_length.restype = ctypes.c_uint
    lib.track_length.argtypes = [ctypes.c_void_p, ctypes.c_uint]

    # get a track
    lib.get.restype = ctypes.c_uint
    lib.get.argtypes = [ctypes.c_void_p, np_dbl_p, ctypes.c_uint]

    # get the internal ID of a track
    lib.get_ID.restype = ctypes.c_uint
    lib.get_ID.argtypes = [ctypes.c_void_p, ctypes.c_uint]

    # get a track, by reference
    lib.get_refs.restype = ctypes.c_uint
    lib.get_refs.argtypes = [ctypes.c_void_p, np_int_vec_p, ctypes.c_uint]

    # get the parent ID (i.e. pre-division)
    lib.get_parent.restype = ctypes.c_uint
    lib.get_parent.argtypes = [ctypes.c_void_p, ctypes.c_uint]

    # get the root ID (i.e. the root ID of this lineage tree)
    lib.get_root.restype = ctypes.c_uint
    lib.get_root.argtypes = [ctypes.c_void_p, ctypes.c_uint]

    # get the ID of any children
    lib.get_children.restype = ctypes.c_uint
    lib.get_children.argtypes = [ctypes.c_void_p, np_int_vec_p, ctypes.c_uint]

    # get the fate of the track
    lib.get_fate.restype = ctypes.c_uint
    lib.get_fate.argtypes = [ctypes.c_void_p, ctypes.c_uint]

    # get the generational depth of the track
    lib.get_generation.restype = ctypes.c_uint
    lib.get_generation.argtypes = [ctypes.c_void_p, ctypes.c_uint]

    # get the kalman filtered position
    lib.get_kalman_mu.restype = ctypes.c_uint
    lib.get_kalman_mu.argtypes = [ctypes.c_void_p, np_dbl_p, ctypes.c_uint]

    # get the kalman covariance
    lib.get_kalman_covar.restype = ctypes.c_uint
    lib.get_kalman_covar.argtypes = [ctypes.c_void_p, np_dbl_p, ctypes.c_uint]

    # get the predicted position at each time step
    lib.get_kalman_pred.restype = ctypes.c_uint
    lib.get_kalman_pred.argtypes = [ctypes.c_void_p, np_dbl_p, ctypes.c_uint]

    # get the label of the object
    lib.get_label.restype = ctypes.c_uint
    lib.get_label.argtypes = [ctypes.c_void_p, np_uint_p, ctypes.c_uint]

    # get the imaging volume
    lib.get_volume.restype = None
    lib.get_volume.argtypes = [ctypes.c_void_p, np_dbl_p]

    # get the imaging volume
    lib.set_volume.restype = None
    lib.set_volume.argtypes = [ctypes.c_void_p, np_dbl_p]

    # return a dummy object by reference
    lib.get_dummy.restype = PyTrackObject
    lib.get_dummy.argtypes = [ctypes.c_void_p, ctypes.c_int]

    # get the number of tracks
    lib.size.restype = ctypes.c_uint
    lib.size.argtypes = [ctypes.c_void_p]

    # get the number of graph edges
    lib.num_edges.restype = ctypes.c_uint
    lib.num_edges.argtypes = [ctypes.c_void_p]

    # get the graph edges
    lib.get_graph_edge.restype = PyGraphEdge
    lib.get_graph_edge.argtypes = [ctypes.c_void_p, ctypes.c_int]

    # calculate the hypotheses
    lib.create_hypotheses.restype = ctypes.c_uint
    lib.create_hypotheses.argtypes = [
        ctypes.c_void_p,
        hypothesis.PyHypothesisParams,
        ctypes.c_uint,
        ctypes.c_uint,
    ]

    # get a hypothesis by ID
    lib.get_hypothesis.restype = hypothesis.Hypothesis
    lib.get_hypothesis.argtypes = [ctypes.c_void_p, ctypes.c_uint]

    # merge following optimisation
    lib.merge.restype = None
    lib.merge.argtypes = [ctypes.c_void_p, np_uint_p, ctypes.c_uint]

    return lib



from contextlib import nullcontext

import numpy as np
import pytest

from btrack import btypes, utils
from btrack.constants import DEFAULT_OBJECT_KEYS, Dimensionality
from btrack.io import objects_from_array

from ._utils import create_test_image, create_test_tracklet


def _example_segmentation_generator():
    for _ in range(10):
        img, centroids = create_test_image()
        yield img


def _validate_centroids(centroids, objects, scale=None):
    """Take a list of objects and validate them agains the ground truth."""

    if centroids is None:
        assert not objects
        return

    if scale is not None:
        centroids = centroids * np.array(scale)

    ndim = centroids.shape[-1]

    obj_as_array = np.array([[obj.z, obj.y, obj.x] for obj in objects])
    if ndim == Dimensionality.TWO:
        obj_as_array = obj_as_array[:, 1:]

    # sort the centroids by axis
    centroids = centroids[
        np.lexsort([centroids[:, dim] for dim in range(ndim)][::-1])
    ]

    # sort the objects
    obj_as_array = obj_as_array[
        np.lexsort([obj_as_array[:, dim] for dim in range(ndim)][::-1])
    ]

    np.testing.assert_equal(obj_as_array, centroids)


def test_segmentation_to_objects_type():
    """Test that btrack objects are returned."""
    img, centroids = create_test_image()
    objects = utils.segmentation_to_objects(img[np.newaxis, ...])
    assert all([isinstance(o, btypes.PyTrackObject) for o in objects])


def test_segmentation_to_objects_type_generator():
    """Test generator as input."""
    generator = _example_segmentation_generator()
    objects = utils.segmentation_to_objects(generator)
    assert all([isinstance(o, btypes.PyTrackObject) for o in objects])


@pytest.mark.parametrize("ndim", [2, 3])
@pytest.mark.parametrize("nobj", [0, 1, 10, 30, 300])
@pytest.mark.parametrize("binary", [True, False])
def test_segmentation_to_objects(ndim, nobj, binary):
    """Test different types of segmentation images."""
    img, centroids = create_test_image(ndim=ndim, nobj=nobj, binary=binary)
    objects = utils.segmentation_to_objects(img[np.newaxis, ...])
    _validate_centroids(centroids, objects)


def test_dask_segmentation_to_objects():
    """Test using a dask array as segmentation input."""
    img, centroids = create_test_image()
    da = pytest.importorskip(
        "dask.array", reason="Dask not installed in pytest environment."
    )
    img = da.from_array(img)
    objects = utils.segmentation_to_objects(img[np.newaxis, ...])
    _validate_centroids(centroids, objects)


@pytest.mark.parametrize("scale", [None, (1.0, 1.0), (1.0, 10.0), (10.0, 1.0)])
def test_segmentation_to_objects_scale(scale):
    """Test anisotropic scaling."""
    img, centroids = create_test_image()
    objects = utils.segmentation_to_objects(img[np.newaxis, ...], scale=scale)
    _validate_centroids(centroids, objects, scale)


@pytest.mark.parametrize("ndim", [2, 3])
@pytest.mark.parametrize("nobj", [0, 1, 10, 30, 300])
def test_assign_class_ID(ndim, nobj):
    """Test mask class_id assignment."""
    img, centroids = create_test_image(ndim=ndim, nobj=nobj, binary=False)
    objects = utils.segmentation_to_objects(
        img[np.newaxis, ...], assign_class_ID=True
    )
    # check that the values match
    for obj in objects:
        centroid = (int(obj.z), int(obj.y), int(obj.x))[-ndim:]
        assert obj.properties["class_id"] == img[centroid]


def test_regionprops():
    """Test using regionprops returns objects with correct property keys."""
    img, centroids = create_test_image()
    properties = (
        "area",
        "axis_major_length",
    )
    objects = utils.segmentation_to_objects(
        img[np.newaxis, ...], properties=properties
    )

    # check that the properties keys match
    for obj in objects:
        assert set(obj.properties.keys()) == set(properties)


@pytest.mark.parametrize("ndim", [2, 3])
def test_intensity_image(default_rng, ndim):
    """Test using an intensity image."""
    img, centroids = create_test_image(ndim=ndim, binary=True)
    intensity_image = img * default_rng.uniform(size=img.shape)
    objects = utils.segmentation_to_objects(
        img[np.newaxis, ...],
        intensity_image=intensity_image[np.newaxis, ...],
        use_weighted_centroid=True,
        properties=("max_intensity",),
    )
    # check that the values match
    for obj in objects:
        centroid = (int(obj.z), int(obj.y), int(obj.x))[-ndim:]
        assert obj.properties["max_intensity"] == intensity_image[centroid]


def test_update_segmentation_2d(test_segmentation_and_tracks):
    """Test relabeling a 2D-segmentation with track ID."""
    in_segmentation, out_segmentation, tracks = test_segmentation_and_tracks
    relabeled = utils.update_segmentation(in_segmentation, tracks)
    assert np.allclose(relabeled, out_segmentation)


@pytest.mark.parametrize("color_by", ["ID", "root", "generation", "fake"])
def test_update_segmentation_2d_colorby(
    test_segmentation_and_tracks, color_by
):
    """Test relabeling a 2D-segmentation with track ID."""
    in_segmentation, out_segmentation, tracks = test_segmentation_and_tracks

    with pytest.raises(ValueError) if color_by == "fake" else nullcontext():
        _ = utils.update_segmentation(
            in_segmentation, tracks, color_by=color_by
        )


def test_update_segmentation_3d(test_segmentation_and_tracks):
    """Test relabeling a 3D-segmentation with track ID."""
    in_segmentation, out_segmentation, tracks = test_segmentation_and_tracks

    in_segmentation = np.broadcast_to(
        in_segmentation[:, None],
        (in_segmentation.shape[0], 5, *in_segmentation.shape[-2:]),
    )

    out_segmentation = np.broadcast_to(
        out_segmentation[:, None],
        (in_segmentation.shape[0], 5, *in_segmentation.shape[-2:]),
    )

    relabeled = utils.update_segmentation(in_segmentation, tracks)
    assert np.allclose(relabeled, out_segmentation)


@pytest.mark.parametrize("ndim", [2, 3])
def test_tracks_to_napari(ndim: int):
    """Test converting from `btrack` Tracklets to a `napari` compatible data
    structure."""

    # make three fake tracks with properties
    track_len = 10
    tracks = [create_test_tracklet(track_len, idx + 1)[0] for idx in range(3)]

    # set up a fake graph
    tracks[0].children = [2, 3]
    tracks[1].parent = 1
    tracks[2].parent = 1

    data, properties, graph = utils.tracks_to_napari(tracks, ndim=ndim)

    # check the data is of the correct shape (ID, T + ndim)
    assert data.shape[-1] == ndim + 2

    # check that the data have the correct values
    track_ids = np.asarray([1] * track_len + [2] * track_len + [3] * track_len)
    np.testing.assert_equal(data[:, 0], track_ids)
    header = ["t"] + ["z", "y", "x"][-ndim:]
    for idx, key in enumerate(header):
        gt_data = np.concatenate([getattr(t, key) for t in tracks])
        np.testing.assert_equal(data[:, idx + 1], gt_data)

    # check the graph
    assert graph == {2: [1], 3: [1]}

    # check the properties keys are correct, note that nD keys are replaced with
    # keys that start with the property key, e.g. `nD` is replaced with `nD-0`
    # and so forth
    for key in tracks[0].properties:
        assert any([k.startswith(key) for k in properties])


@pytest.mark.parametrize("ndim", [1, 4])
def test_tracks_to_napari_incorrect_ndim(ndim: int):
    """Test that providing incorrect dimensions to `tracks_to_napari` raises a
    `ValueError`."""
    # make three fake tracks with properties
    track_len = 10
    tracks = [create_test_tracklet(track_len, idx + 1)[0] for idx in range(3)]

    with pytest.raises(ValueError):
        data, properties, graph = utils.tracks_to_napari(tracks, ndim=ndim)


@pytest.mark.parametrize("ndim", [2, 3])
def test_tracks_to_napari_ndim_inference(ndim: int):
    """Test inferring the correct dimensions from track data when using
    `tracks_to_napari`."""

    # make a fake track with n dimensions
    track_len = 10
    tracks = [create_test_tracklet(track_len, 1, ndim=ndim)[0]]
    data, _, _ = utils.tracks_to_napari(tracks, ndim=None)

    # check the data is of the correct shape (ID, T + ndim)
    assert data.shape[-1] == ndim + 2


def test_objects_from_array(test_objects):
    """Test creation of a list of objects from a numpy array."""

    obj_arr = np.stack(
        [
            [getattr(obj, k) for k in DEFAULT_OBJECT_KEYS]
            for obj in test_objects
        ],
        axis=0,
    )

    obj_from_arr = objects_from_array(obj_arr)

    assert obj_arr.shape[0] == len(test_objects)
    assert obj_arr.shape[-1] == len(DEFAULT_OBJECT_KEYS)

    assert len(obj_from_arr) == len(test_objects)

    for test_obj, obj in zip(test_objects, obj_from_arr):
        assert isinstance(obj, btypes.PyTrackObject)
        for key in DEFAULT_OBJECT_KEYS:
            assert getattr(test_obj, key) == getattr(obj, key)



import enum
import os

BTRACK_PATH = os.path.dirname(os.path.abspath(__file__))

MAX_SEARCH_RADIUS = 100
DEFAULT_LOW_PROBABILITY = -1e5
MAX_LOST = 5
PROB_NOT_ASSIGN = 0.1
DEBUG = True
EXPORT_FORMATS = frozenset([".json", ".mat", ".hdf5"])
VOLUME = ((0, 1024), (0, 1024), (-1e5, 1e5))
GLPK_OPTIONS = {"tm_lim": 60_000}


DEFAULT_OBJECT_KEYS = ["t", "x", "y", "z", "label"]
DEFAULT_EXPORT_PROPERTIES = [
    "ID",
    "t",
    "x",
    "y",
    "z",
    "parent",
    "root",
    "state",
    "generation",
    "dummy",
]


@enum.unique
class Errors(enum.Enum):
    SUCCESS = 900
    EMPTY_QUEUE = 901
    NO_TRACKS = 902
    NO_USEABLE_FRAMES = 903
    TRACK_EMPTY = 904
    INCORRECT_MOTION_MODEL = 905
    MAX_LOST_OUT_OF_RANGE = 906
    ACCURACY_OUT_OF_RANGE = 907
    PROB_NOT_ASSIGN_OUT_OF_RANGE = 908
    NOT_DEFINED = 909
    NO_ERROR = 910


@enum.unique
class Fates(enum.Enum):
    FALSE_POSITIVE = 0
    INITIALIZE = 1
    TERMINATE = 2
    LINK = 3
    DIVIDE = 4
    APOPTOSIS = 5
    MERGE = 6
    EXTRUDE = 7
    INITIALIZE_BORDER = 10
    INITIALIZE_FRONT = 11
    INITIALIZE_LAZY = 12
    TERMINATE_BORDER = 20
    TERMINATE_BACK = 21
    TERMINATE_LAZY = 22
    DEAD = 666
    UNDEFINED = 999


@enum.unique
class States(enum.Enum):
    INTERPHASE = 0
    PROMETAPHASE = 1
    METAPHASE = 2
    ANAPHASE = 3
    APOPTOSIS = 4
    NULL = 5
    DUMMY = 99


@enum.unique
class BayesianUpdates(enum.Enum):
    EXACT = 0
    APPROXIMATE = 1
    CUDA = 2


@enum.unique
class BayesianUpdateFeatures(enum.Enum):
    MOTION = 0b0000001
    VISUAL = 0b0000010


class Dimensionality(enum.IntEnum):
    TWO: int = 2
    THREE: int = 3
    FOUR: int = 4



try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown"

import logging

from btrack.napari import constants, main

__all__ = [
    "constants",
    "main",
]



import json
import logging
import os
from pathlib import Path
from typing import List, Optional

import numpy as np
from pydantic import BaseModel, conlist, validator

from btrack import _version

from . import constants
from .btypes import ImagingVolume
from .models import HypothesisModel, MotionModel, ObjectModel
from .utils import read_hypothesis_model, read_motion_model, read_object_model

# get the logger instance
logger = logging.getLogger(__name__)


class TrackerConfig(BaseModel):
    """Configuration for `BayesianTracker`.

    Parameters
    ----------
    name : str
        A name identifier for the model.
    version : str
        A string representing the version of `btrack` used.
    verbose : bool
        A flag to set the verbosity level while logging the output.
    motion_model : Optional[MotionModel]
        The `btrack` motion model. See `models.MotionModel` for more details.
    object_model : Optional[ObjectModel]
        The `btrack` object model. See `models.ObjectModel` for more details.
    hypothesis_model : Optional[HypothesisModel]
        The `btrack` hypothesis model. See `models.HypothesisModel` for more
        details.
    max_search_radius : float
        The maximum search radius of the algorithm in isotropic units of the
        data. Should be greater than zero.
    return_kalman : bool
        Flag to request the Kalman debug info when returning tracks.
    volume : Optional[ImagingVolume]
        The imaging volume as [(xlo, xhi), ..., (zlo, zhi)]. See
        :py:meth:`btrack.btypes.ImagingVolume` for more details.
    update_method : constants.BayesianUpdates
        The method to perform the bayesian updates during tracklet linking.

            * BayesianUpdates.EXACT
                Use the exact Bayesian update method. Can be slow for systems
                with many objects.
            * BayesianUpdates.APPROXIMATE
                Use the approximate Bayesian update method. Useful for systems
                with may objects.
            * BayesianUpdates.CUDA
                Use the CUDA implementation of the Bayesian update method. Not
                currently implemented.
    optimizer_options: dict
        Additional options to pass to the optimizer. See `cvxopt.glpk` for more
        details of options that can be passed.
    features : list
        A list of feature names to be used during tracking updates. These must
        correspond to named features found in the properties of each
        `:py:class:btrack.btypes.PyTrackObject` in the dataset.
    tracking_updates : list
        A list of features to be used for tracking, such as MOTION or VISUAL.
        Must have at least one entry.

    Notes
    -----
    TODO(arl): add more validation to parameters.
    """

    name: str = "Default"
    version: str = _version.version
    verbose: bool = False
    motion_model: Optional[MotionModel] = None
    object_model: Optional[ObjectModel] = None
    hypothesis_model: Optional[HypothesisModel] = None
    max_search_radius: float = constants.MAX_SEARCH_RADIUS
    return_kalman: bool = False
    store_candidate_graph: bool = False
    volume: Optional[ImagingVolume] = None
    update_method: constants.BayesianUpdates = constants.BayesianUpdates.EXACT
    optimizer_options: dict = constants.GLPK_OPTIONS
    features: List[str] = []
    tracking_updates: conlist(
        constants.BayesianUpdateFeatures,
        min_items=1,
        max_items=len(constants.BayesianUpdateFeatures),
    ) = [
        constants.BayesianUpdateFeatures.MOTION,
    ]

    @validator("volume", pre=True, always=True)
    def _parse_volume(cls, v):
        if isinstance(v, tuple):
            return ImagingVolume(*v)
        return v

    @validator("tracking_updates", pre=True, always=True)
    def _parse_tracking_updates(cls, v):
        _tracking_updates = v
        if all(isinstance(k, str) for k in _tracking_updates):
            _tracking_updates = [
                constants.BayesianUpdateFeatures[k.upper()]
                for k in _tracking_updates
            ]
        _tracking_updates = list(set(_tracking_updates))
        return _tracking_updates

    class Config:
        arbitrary_types_allowed = True
        validate_assignment = True
        json_encoders = {
            np.ndarray: lambda x: x.ravel().tolist(),
        }


def load_config(filename: os.PathLike) -> TrackerConfig:
    """Load a tracker configuration from a file.

    Parameters
    ----------
    filename : os.PathLike
        The filename to load the file.

    Returns
    -------
    cfg : TrackerConfig
        The tracker configuration.
    """
    logger.info(f"Loading configuration file: {filename}")
    filename = Path(filename)

    with open(filename, "r") as json_file:
        json_data = json.load(json_file)

    if "TrackerConfig" in json_data:
        cfg = _load_legacy_config(json_data)
    else:
        cfg = _load_config(json_data)

    assert cfg.motion_model is not None
    return cfg


def _load_legacy_config(json_data: dict) -> TrackerConfig:
    """Load a legacy config file."""
    config = json_data["TrackerConfig"]

    t_config = {
        "motion_model": read_motion_model(config),
        "object_model": read_object_model(config),
        "hypothesis_model": read_hypothesis_model(config),
    }

    return TrackerConfig(**t_config)


def _load_config(json_data: dict) -> TrackerConfig:
    """Load a new style config from a JSON file."""
    return TrackerConfig(**json_data)


def save_config(filename: os.PathLike, cfg: TrackerConfig) -> None:
    """Save the config to a JSON file.

    Parameters
    ----------
    filename : os.PathLike
        The filename to save the configuration file.
    cfg : TrackerConfig
        The tracker configuration to save.
    """

    with open(filename, "w") as json_file:
        json_data = json.loads(cfg.json())
        json.dump(json_data, json_file, indent=2, separators=(",", ": "))



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from btrack.config import TrackerConfig

import copy
import os
from dataclasses import dataclass, field

import numpy as np

import btrack
from btrack import datasets

__all__ = [
    "create_default_configs",
]


@dataclass
class Sigmas:
    """Values to scale TrackerConfig MotionModel matrices by.

    Args:
        P: Scaling factor for the matrix ``P`` - the error in initial estimates.
        G: Scaling factor for the matrix ``G`` - the error in the MotionModel process.
        R: Scaling factor for the matrix ``R`` - the error in the measurements.

    """

    P: float
    G: float
    R: float

    def __getitem__(self, matrix_name):
        return self.__dict__[matrix_name]

    def __setitem__(self, matrix_name, sigma):
        if matrix_name not in self.__dict__.keys():
            _msg = f"Unknown matrix name '{matrix_name}'"
            raise ValueError(_msg)
        self.__dict__[matrix_name] = sigma

    def __iter__(self):
        yield from self.__dict__.keys()


@dataclass
class UnscaledTrackerConfig:
    """Convert TrackerConfig MotionModel matrices from scaled to unscaled.

    This is needed because TrackerConfig stores "scaled" matrices, i.e. it
    doesn't store sigma and the "unscaled" MotionModel matrices separately.

    Args:
        filename: name of the json file containing the TrackerConfig to load.

    Attributes:
        tracker_config: unscaled configuration based on the config in ``filename``.
        sigmas: scaling factors to apply to the unscaled MotionModel matrices of
            ``tracker_config``.

    """

    filename: os.PathLike
    tracker_config: TrackerConfig = field(init=False)
    sigmas: Sigmas = field(init=False)

    def __post_init__(self):
        """Create the TrackerConfig and un-scale the MotionModel indices"""

        config = btrack.config.load_config(self.filename)
        self.tracker_config, self.sigmas = self._unscale_config(config)

    def _unscale_config(
        self, config: TrackerConfig
    ) -> tuple[TrackerConfig, Sigmas]:
        """Convert the matrices of a scaled TrackerConfig MotionModel to unscaled."""

        P_sigma = np.max(config.motion_model.P)
        config.motion_model.P /= P_sigma

        R_sigma = np.max(config.motion_model.R)
        config.motion_model.R /= R_sigma

        # Use only G, not Q. If we use both G and Q, then Q_sigma must be updated
        # when G_sigma is, and vice-versa
        # Instead, use G if it exists. If not, determine G from Q, which we can
        # do because Q = G.T @ G
        if config.motion_model.G is None:
            config.motion_model.G = config.motion_model.Q.diagonal() ** 0.5
        G_sigma = np.max(config.motion_model.G)
        config.motion_model.G /= G_sigma

        sigmas = Sigmas(
            P=P_sigma,
            G=G_sigma,
            R=R_sigma,
        )

        return config, sigmas

    def scale_config(self) -> TrackerConfig:
        """Create a new TrackerConfig with scaled MotionModel matrices"""

        # Create a copy so that config values stay in sync with widget values
        scaled_config = copy.deepcopy(self.tracker_config)
        scaled_config.motion_model.P *= self.sigmas.P
        scaled_config.motion_model.R *= self.sigmas.R
        scaled_config.motion_model.G *= self.sigmas.G
        scaled_config.motion_model.Q = (
            scaled_config.motion_model.G.T @ scaled_config.motion_model.G
        )

        return scaled_config


@dataclass
class TrackerConfigs:
    """Store all loaded TrackerConfig configurations.

    Will load ``btrack``'s default 'cell' and 'particle' configurations on
    initialisation.

    Attributes:
        configs: dictionary of loaded configurations. The name of the config (
            TrackerConfig.name) is used as the key.
        current_config: the currently-selected configuration.

    """

    configs: dict[str, UnscaledTrackerConfig] = field(default_factory=dict)
    current_config: str = field(init=False)

    def __post_init__(self):
        """Add the default cell and particle configs."""

        self.add_config(
            filename=datasets.cell_config(),
            name="cell",
            overwrite=False,
        )
        self.add_config(
            filename=datasets.particle_config(),
            name="particle",
            overwrite=False,
        )

        self.current_config = "cell"

    def __getitem__(self, config_name):
        return self.configs[config_name]

    def add_config(
        self,
        filename,
        overwrite,
        name=None,
    ) -> str:
        """Load a TrackerConfig and add it to the store."""

        config = UnscaledTrackerConfig(filename)
        config_name = config.tracker_config.name if name is None else name
        config.tracker_config.name = config_name

        # TODO: Make the combobox editable so config names can be changed within the GUI
        if config_name in self.configs and not overwrite:
            _msg = (
                f"Config '{config_name}' already exists - "
                "config names must be unique."
            )
            raise ValueError(_msg)

        self.configs[config_name] = config

        return config_name


def create_default_configs() -> TrackerConfigs:
    """Create a set of default configurations."""

    # TrackerConfigs automatically loads default cell and particle configs
    return TrackerConfigs()



"""
This module contains variables that are used throughout the
napari plugin of the btrack package.
"""

HYPOTHESES = [
    "P_FP",
    "P_init",
    "P_term",
    "P_link",
    "P_branch",
    "P_dead",
    "P_merge",
]

HYPOTHESIS_SCALING_FACTORS = [
    "lambda_time",
    "lambda_dist",
    "lambda_link",
    "lambda_branch",
]

HYPOTHESIS_THRESHOLDS = [
    "theta_dist",
    "theta_time",
    "dist_thresh",
    "time_thresh",
    "apop_thresh",
    "relax",
]



#!/usr/bin/env python
# -------------------------------------------------------------------------------
# Name:     btrack
# Purpose:  A multi object tracking library, specifically used to reconstruct
#           tracks in crowded fields. Here we use a probabilistic network of
#           information to perform the trajectory linking. This method uses
#           positional and visual information for track linking.
#
# Authors:  Alan R. Lowe (arl) a.lowe@ucl.ac.uk
#
# License:  See LICENSE.md
#
# Created:  14/08/2014
# -------------------------------------------------------------------------------

from __future__ import annotations

import ctypes
from collections import OrderedDict
from typing import Any, Dict, List, NamedTuple, Optional, Tuple

import numpy as np

from . import constants

__all__ = ["PyTrackObject", "Tracklet"]


class ImagingVolume(NamedTuple):
    x: Tuple[float, float]
    y: Tuple[float, float]
    z: Optional[Tuple[float, float]] = None

    @property
    def ndim(self) -> int:
        """Infer the dimensionality from the volume."""
        return (
            constants.Dimensionality.TWO
            if self.z is None
            else constants.Dimensionality.THREE
        )


class PyTrackObject(ctypes.Structure):
    """The base `btrack` track object.

    Primitive class to store information about an object. Essentially a single
    object in a field of view, with some member variables to keep track of data
    associated with an object.

    Parameters
    ----------
    ID : int
        The unique ID of the object.
    x : float
        The x coordinate.
    y : float
        The y coordinate.
    z : float
        The z coordinate.
    t : int
        The timestamp.
    dummy: bool
        Flag for whether the objects is real or a dummy (inserted by the
        tracker when no observation can be linked).
    states : int
        The number of states of the object. This corresponds to the number of
        possible labels.
    label : int
        The label of the object.
    features : array
        A vector of feature values.
    n_features : int
        The length of the feature vector.

    Attributes
    ----------
    properties : Dict[str, Union[int, float]]
        Dictionary of properties associated with this object.
    state : constants.States
        A state label for the object. See `constants.States`

    Notes
    -----
    stackoverflow.com/questions/23329663/access-np-array-in-ctypes-struct

    """

    _fields_ = [
        ("ID", ctypes.c_long),
        ("x", ctypes.c_double),
        ("y", ctypes.c_double),
        ("z", ctypes.c_double),
        ("t", ctypes.c_uint),
        ("dummy", ctypes.c_bool),
        ("states", ctypes.c_uint),
        ("label", ctypes.c_int),
        ("n_features", ctypes.c_int),
        ("features", ctypes.POINTER(ctypes.c_double)),
    ]

    def __init__(self):
        super().__init__()
        self.dummy = False
        self.label = constants.States.NULL.value
        self.states = len(constants.States)
        self.n_features = 0
        self._properties = {}

    @property
    def properties(self) -> Dict[str, Any]:
        if self.dummy:
            return {}
        return self._properties

    @properties.setter
    def properties(self, properties: Dict[str, Any]):
        """Set the object properties."""
        self._properties.update(properties)

    @property
    def state(self) -> constants.States:
        return constants.States(self.label)

    def set_features(self, keys: List[str]) -> None:
        """Set features to be used by the tracking update."""

        if not keys:
            self.n_features = 0
            return

        if not all(k in self.properties for k in keys):
            missing_features = list(
                set(keys).difference(set(self.properties.keys()))
            )
            raise KeyError(f"Feature(s) missing: {missing_features}.")

        # store a reference to the numpy array so that Python maintains
        # ownership of the memory allocated to the numpy array
        self._features = np.concatenate(
            [np.asarray(self.properties[k]).ravel() for k in keys], axis=-1
        ).astype(np.float64)

        # NOTE(arl): do we want to normalise the features here???
        # self._features = features / np.linalg.norm(features)
        self.features = np.ctypeslib.as_ctypes(self._features)
        self.n_features = len(self._features)

    def to_dict(self) -> Dict[str, Any]:
        """Return a dictionary of the fields and their values."""
        node = {
            k: getattr(self, k)
            for k, _ in PyTrackObject._fields_
            if k not in ("features", "n_features")
        }
        node.update(self.properties)
        return node

    @staticmethod
    def from_dict(properties: Dict[str, Any]) -> PyTrackObject:
        """Build an object from a dictionary."""
        obj = PyTrackObject()
        fields = {k: kt for k, kt in PyTrackObject._fields_}
        attr = [k for k in fields if k in properties]
        for key in attr:
            new_data = properties[key]

            # fix for implicit type conversion
            if key in ("ID", "t", "states", "label"):
                setattr(obj, key, int(new_data))
            elif key in ("dummy",):
                setattr(obj, key, bool(new_data))
            else:
                setattr(obj, key, float(new_data))

        # we can add any extra details to the properties dictionary
        obj.properties = {
            k: v for k, v in properties.items() if k not in fields.keys()
        }
        return obj

    def __repr__(self):
        return self.to_dict().__repr__()

    def _repr_html_(self):
        return _pandas_html_repr(self)


class PyTrackingInfo(ctypes.Structure):
    """Primitive class to store information about the tracking output.

    Parameters
    ----------
    error : int
        Error code from the tracker. See `constants.Errors` for definitions.
    n_tracks : int
        Total number of tracks initialised during tracking.
    n_active : int
        Number of active tracks.
    n_conflicts : int
        Number of conflicts.
    n_lost : int
        Number of lost tracks.
    t_update_belief : float
        Time to update belief matrix in ms.
    t_update_link : float
        Time to update links in ms.
    t_total_time : float
        Total time to track objects.
    p_link : float
        Typical probability of association.
    p_lost : float
        Typical probability of losing track.
    complete : bool
        Flag denoting that the tracking is complete.

    Notes
    -----
    TODO(arl): should update to give more useful statistics, perhaps
    histogram of probabilities and timings.

    """

    _fields_ = [
        ("error", ctypes.c_uint),
        ("n_tracks", ctypes.c_uint),
        ("n_active", ctypes.c_uint),
        ("n_conflicts", ctypes.c_uint),
        ("n_lost", ctypes.c_uint),
        ("t_update_belief", ctypes.c_float),
        ("t_update_link", ctypes.c_float),
        ("t_total_time", ctypes.c_float),
        ("p_link", ctypes.c_float),
        ("p_lost", ctypes.c_float),
        ("complete", ctypes.c_bool),
    ]

    def to_dict(self) -> Dict[str, Any]:
        """Return a dictionary of the statistics"""
        # TODO(arl): make this more readable by converting seconds, ms
        # and interpreting error messages?
        stats = {k: getattr(self, k) for k, typ in PyTrackingInfo._fields_}
        return stats

    @property
    def tracker_active(self) -> bool:
        """Return the current status."""
        no_error = constants.Errors(self.error) == constants.Errors.NO_ERROR
        return no_error and not self.complete


class PyGraphEdge(ctypes.Structure):
    """A structure defining an edge in the association graph. This is derived
    from the Bayesian belief matrix in the initial step of the tracking
    algorithm.

    Parameters
    ----------
    source : int
        A reference to a source object.
    target : int
        A reference to a target object.
    score : float
        The posterior probability of linking the target object to the source.

    Notes
    -----
    This structure does not guarantee that the target timestamp is *after* the
    source timestamp, we just assume that the tracker has done it's job.
    """

    _fields_ = [
        ("source", ctypes.c_long),
        ("target", ctypes.c_long),
        ("score", ctypes.c_double),
        ("edge_type", ctypes.c_uint),
    ]

    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary describing the edge."""
        edge = {k: getattr(self, k) for k, _ in PyGraphEdge._fields_}
        return edge


class Tracklet:
    """A `btrack` Tracklet object used to store track information.

    Parameters
    ----------
    ID : int
        A unique integer identifier for the tracklet.
    data : list[PyTrackObject]
        The objects linked together to form the track.
    parent : int
        The identifiers of the parent track(s).
    children : list
        The identifiers of the child tracks.
    fate : constants.Fates, default = constants.Fates.UNDEFINED
        An enumerated type describing the fate of the track.

    Attributes
    ----------
    x : list[float]
        The list of x positions.
    y : list[float]
        The list of y positions.
    z : list[float]
        The list of z positions.
    t : list[float]
        The list of timestamps.
    dummy : list[bool]
        A list specifying which objects are dummy objects inserted by the tracker.
    parent : int, list
        The identifiers of the parent track(s).
    refs : list[int]
        Returns a list of :py:class:`btrack.btypes.PyTrackObject` identifiers
        used to build the track. Useful for indexing back into the original
        data, e.g. table of localizations or h5 file.
    label : list[str]
        Return the label of each object in the track.
    state : list[int]
        Return the numerical label of each object in the track.
    softmax : list[float]
        If defined, return the softmax score for the label of each object in the
        track.
    properties : Dict[str, np.ndarray]
        Return a dictionary of track properties derived from
        :py:class:`btrack.btypes.PyTrackObject` properties.
    root : int,
        The identifier of the root ID if a branching tree (ie cell division).
    is_root : boole
        Flag to denote root track.
    is_leaf : bool
        Flag to denote leaf track.
    start : int, float
        First time stamp of track.
    stop : int, float
        Last time stamp of track.
    kalman : np.ndarray
        Return the complete output of the kalman filter for this track. Note,
        that this may not have been returned while from the tracker. See
        :py:attr:`btrack.BayesianTracker.return_kalman` for more details.
    LBEP : list
        An LBEP representation of the track.


    Notes
    -----
    Tracklet object for storing and updating linked lists of track objects.
    Forms the data structure for an individual tracklet. Track 'fates' are the
    selected hypotheses after optimization. Defined in constants.Fates. Intrinsic
    properties can be accesses as attributes, e.g: track.x returns the track
    x values.
    """

    def __init__(
        self,
        ID: int,
        data: List[PyTrackObject],
        *,
        parent: Optional[int] = None,
        children: Optional[List[int]] = None,
        fate: constants.Fates = constants.Fates.UNDEFINED,
    ):
        assert all([isinstance(o, PyTrackObject) for o in data])

        self.ID = ID
        self._data = data
        self._kalman = None

        self.root = None
        self.parent = parent
        self.children = children if children is not None else []
        self.type = None
        self.fate = fate
        self.generation = 0

    def __len__(self):
        return len(self._data)

    def __repr__(self):
        return self.to_dict().__repr__()

    def _repr_html_(self):
        return _pandas_html_repr(self)

    @property
    def properties(self) -> Dict[str, np.ndarray]:
        """Return the properties of the objects."""
        # find the set of keys, then grab the properties
        keys = set()
        for obj in self._data:
            keys.update(obj.properties.keys())

        # work out the shapes of the properties by finding the first object that
        # is not a dummy and returning the shape of the property, we can use
        # this to fill the properties array with NaN for dummy objects
        property_shapes = {
            k: next(
                (
                    np.asarray(o.properties[k]).shape
                    for o in self._data
                    if not o.dummy
                ),
                None,
            )
            for k in keys
        }

        # set the properties, replacing missing values with a NaN
        properties = {
            k: [
                o.properties[k]
                if k in o.properties
                else np.full(property_shapes[k], np.nan)
                for o in self._data
            ]
            for k in keys
        }

        # validate the track properties
        for k, v in properties.items():
            if len(v) != len(self):
                raise ValueError(
                    "The number of properties and track objects must be equal."
                )
            # ensure the property values are a numpy array
            if not isinstance(v, np.ndarray):
                properties[k] = np.asarray(v)

        return properties

    @properties.setter
    def properties(self, properties: Dict[str, np.ndarray]):
        """Store properties associated with this Tracklet."""
        # TODO(arl): this will need to set the object properties
        pass

    def __getitem__(self, attr: str):
        assert isinstance(attr, str)
        try:
            return getattr(self, attr)
        except AttributeError:
            return self.properties[attr]

    @property
    def x(self) -> list:
        return [o.x for o in self._data]

    @property
    def y(self) -> list:
        return [o.y for o in self._data]

    @property
    def z(self) -> list:
        return [o.z for o in self._data]

    @property
    def t(self) -> list:
        return [o.t for o in self._data]

    @property
    def dummy(self) -> list:
        return [o.dummy for o in self._data]

    @property
    def refs(self) -> list:
        return [o.ID for o in self._data]

    @property
    def start(self) -> list:
        return self.t[0]

    @property
    def stop(self) -> list:
        return self.t[-1]

    @property
    def label(self) -> list:
        return [o.state.name for o in self._data]

    @property
    def state(self) -> list:
        return [o.state.value for o in self._data]

    @property
    def softmax(self) -> list:
        return [o.probability for o in self._data]

    @property
    def is_root(self) -> bool:
        return (
            self.parent == 0 or self.parent is None or self.parent == self.ID
        )

    @property
    def is_leaf(self) -> bool:
        return not self.children

    @property
    def kalman(self) -> np.ndarray:
        return self._kalman

    @kalman.setter
    def kalman(self, data: np.ndarray) -> None:
        assert isinstance(data, np.ndarray)
        self._kalman = data

    def mu(self, index: int) -> np.ndarray:
        """Return the Kalman filter mu. Note that we are only returning the mu
        for the positions (e.g. 3x1)."""
        return self.kalman[index, 1:4].reshape(3, 1)

    def covar(self, index: int) -> np.ndarray:
        """Return the Kalman filter covariance matrix. Note that we are
        only returning the covariance matrix for the positions (e.g. 3x3)."""
        return self.kalman[index, 4:13].reshape(3, 3)

    def predicted(self, index: int) -> np.ndarray:
        """Return the motion model prediction for the given timestep."""
        return self.kalman[index, 13:].reshape(3, 1)

    def to_dict(
        self, properties: list = constants.DEFAULT_EXPORT_PROPERTIES
    ) -> Dict[str, Any]:
        """Return a dictionary of the tracklet which can be used for JSON
        export. This is an ordered dictionary for nicer JSON output.
        """
        trk_tuple = tuple([(p, getattr(self, p)) for p in properties])
        data = OrderedDict(trk_tuple)
        data.update(self.properties)
        return data

    def to_array(
        self, properties: list = constants.DEFAULT_EXPORT_PROPERTIES
    ) -> np.ndarray:
        """Return a representation of the trackled as a numpy array."""
        data = self.to_dict(properties)
        tmp_track = []
        for values in data.values():
            np_values = np.asarray(values)
            if np_values.size == 1:
                np_values = np.tile(np_values, len(self))
            np_values = np.reshape(np_values, (len(self), -1))
            tmp_track.append(np_values)

        tmp_track = np.concatenate(tmp_track, axis=-1)
        assert tmp_track.shape[0] == len(self)
        assert tmp_track.ndim == constants.Dimensionality.TWO
        return tmp_track.astype(np.float32)

    def in_frame(self, frame: int) -> bool:
        """Return true or false as to whether the track is in the frame."""
        return self.t[0] <= frame and self.t[-1] >= frame

    def trim(self, frame: int, tail: int = 75) -> Tracklet:
        """Trim the tracklet and return one with the trimmed data."""
        d = [o for o in self._data if o.t <= frame and o.t >= frame - tail]
        return Tracklet(self.ID, d)

    def LBEP(self) -> Tuple[int]:
        """Return an LBEP table summarising the track."""
        return (
            self.ID,
            self.start,
            self.stop,
            self.parent,
            self.root,
            self.generation,
        )


def _pandas_html_repr(obj):
    """Prepare data for HTML representation in a notebook."""
    try:
        import pandas as pd
    except ImportError:
        return (
            "<b>Install pandas for nicer, tabular rendering.</b> <br>"
            + obj.__repr__()
        )

    obj_as_dict = obj.to_dict()

    # now try to process for display in the notebook
    n_items = len(obj) if hasattr(obj, "__len__") else 1

    for k, v in obj_as_dict.items():
        if not isinstance(v, (list, np.ndarray)):
            obj_as_dict[k] = [v] * n_items
        elif isinstance(v, np.ndarray):
            ndim = 0 if n_items == 1 else 1
            if v.ndim > ndim:
                obj_as_dict[k] = [f"{v.shape[ndim:]} array"] * n_items

    return pd.DataFrame.from_dict(obj_as_dict).to_html()



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from magicgui.widgets import Container

import json
from unittest.mock import patch

import napari
import numpy as np
import numpy.typing as npt
import pytest

import btrack
import btrack.napari
import btrack.napari.main
from btrack import datasets
from btrack.datasets import cell_config, particle_config

OLD_WIDGET_LAYERS = 1
NEW_WIDGET_LAYERS = 2


def test_add_widget(make_napari_viewer):
    """Checks that the track widget can be added inside a dock widget."""

    viewer = make_napari_viewer()
    num_dw = len(list(viewer.window._dock_widgets))
    viewer.window.add_plugin_dock_widget(
        plugin_name="btrack",
        widget_name="Track",
    )

    assert len(list(viewer.window._dock_widgets)) == num_dw + 1


@pytest.fixture
def track_widget(make_napari_viewer) -> Container:
    """Provides an instance of the track widget to test"""
    make_napari_viewer()  # make sure there is a viewer available
    return btrack.napari.main.create_btrack_widget()


@pytest.mark.parametrize("config", [cell_config(), particle_config()])
def test_config_to_widgets_round_trip(track_widget, config):
    """Tests that going back and forth between
    config objects and widgets works as expected.
    """

    expected_config = btrack.config.load_config(config).json()

    unscaled_config = btrack.napari.config.UnscaledTrackerConfig(config)
    btrack.napari.sync.update_widgets_from_config(
        unscaled_config, track_widget
    )
    btrack.napari.sync.update_config_from_widgets(
        unscaled_config, track_widget
    )

    actual_config = unscaled_config.scale_config().json()

    # use json.loads to avoid failure in string comparison because e.g "100.0" != "100"
    assert json.loads(actual_config) == json.loads(expected_config)


def test_save_button(track_widget):
    """Tests that clicking the save configuration button
    triggers a call to btrack.config.save_config with expected arguments.
    """

    unscaled_config = btrack.napari.config.UnscaledTrackerConfig(cell_config())
    unscaled_config.tracker_config.name = (
        "cell"  # this is done in in the gui too
    )
    expected_config = unscaled_config.scale_config().json()

    with patch(
        "btrack.napari.widgets.save_path_dialogue_box"
    ) as save_path_dialogue_box:
        save_path_dialogue_box.return_value = "user_config.json"
        track_widget.save_config_button.clicked()

    actual_config = btrack.config.load_config("user_config.json").json()

    # use json.loads to avoid failure in string comparison because e.g "100.0" != "100"
    assert json.loads(expected_config) == json.loads(actual_config)


def test_load_config(track_widget):
    """Tests that another TrackerConfig can be loaded and made the current config."""

    # this is set to be 'cell' rather than 'Default'
    original_config_name = track_widget.config.current_choice

    with patch(
        "btrack.napari.widgets.load_path_dialogue_box"
    ) as load_path_dialogue_box:
        load_path_dialogue_box.return_value = cell_config()
        track_widget.load_config_button.clicked()

    # We didn't override the name, so it should be 'Default'
    new_config_name = track_widget.config.current_choice

    assert track_widget.config.value == "Default"
    assert new_config_name != original_config_name


def test_reset_button(track_widget):
    """Tests that clicking the reset button restores the default config values"""

    original_max_search_radius = track_widget.max_search_radius.value
    original_relax = track_widget.relax.value

    # change some widget values
    track_widget.max_search_radius.value += 10
    track_widget.relax.value = not track_widget.relax

    # click reset button - restores defaults of the currently-selected base config
    track_widget.reset_button.clicked()

    new_max_search_radius = track_widget.max_search_radius.value
    new_relax = track_widget.relax.value

    assert new_max_search_radius == original_max_search_radius
    assert new_relax == original_relax


@pytest.fixture
def simplistic_tracker_outputs() -> (
    tuple[npt.NDArray, dict[str, npt.NDArray], dict[int, list]]
):
    """Provides simplistic return values of a btrack run.

    They have the correct types and dimensions, but contain zeros.
    Useful for mocking the tracker.
    """
    n, d = 10, 3
    data = np.zeros((n, d + 1))
    properties = {"some_property": np.zeros(n)}
    graph = {0: [0]}
    return data, properties, graph


def test_run_button(track_widget, simplistic_tracker_outputs):
    """Tests that clicking the run button calls run_tracker,
    and that the napari viewer has an additional tracks layer after running.
    """
    with patch("btrack.napari.main._run_tracker") as run_tracker:
        run_tracker.return_value = simplistic_tracker_outputs
        segmentation = datasets.example_segmentation()
        track_widget.viewer.add_labels(segmentation)
        assert len(track_widget.viewer.layers) == OLD_WIDGET_LAYERS
        track_widget.call_button.clicked()
    assert run_tracker.called
    assert len(track_widget.viewer.layers) == NEW_WIDGET_LAYERS
    assert isinstance(track_widget.viewer.layers[-1], napari.layers.Tracks)



from btrack import datasets


def test_pooch_registry():
    """
    Test that `pooch` registry is up to date with remote version.
    This will fail if the remote file does not match the hash hard-coded
    in btrack.datasets.
    """
    registry_file = datasets._remote_registry()  # noqa: F841



import os
from pathlib import Path

import numpy as np
import pytest

import btrack

from ._utils import (
    create_test_object,
    create_test_properties,
    full_tracker_example,
    simple_tracker_example,
)


def test_hdf5_write(hdf5_file_path, test_objects):
    """Test writing an HDF5 file with some objects."""
    # now try to read those objects and compare with those used to write
    with btrack.io.HDF5FileHandler(hdf5_file_path, "r") as h:
        objects_from_file = h.objects

    properties = ["x", "y", "z", "t", "label", "ID"]

    for orig, read in zip(test_objects, objects_from_file):
        for p in properties:
            # use all close, since h5 file stores in float32 default
            np.testing.assert_allclose(getattr(orig, p), getattr(read, p))


def test_hdf5_write_with_properties(hdf5_file_path):
    """Test writing an HDF5 file with some objects with additional properties."""

    objects = []
    for i in range(10):
        obj, _ = create_test_object(test_id=i)
        obj.properties = create_test_properties()
        objects.append(obj)

    with btrack.io.HDF5FileHandler(hdf5_file_path, "w") as h:
        h.write_objects(objects)

    # now try to read those objects and compare with those used to write
    with btrack.io.HDF5FileHandler(hdf5_file_path, "r") as h:
        objects_from_file = h.objects

    extra_props = list(create_test_properties().keys())

    properties = ["x", "y", "z", "t", "label", "ID"]

    for orig, read in zip(objects, objects_from_file):
        for p in properties:
            # use all close, since h5 file stores in float32 default
            np.testing.assert_allclose(getattr(orig, p), getattr(read, p))
        for p in extra_props:
            np.testing.assert_allclose(orig.properties[p], read.properties[p])


@pytest.mark.parametrize("export_format", ["", ".csv", ".h5"])
def test_tracker_export(tmp_path, export_format):
    """Test that file export works using the `export_delegator`."""

    tracker, _ = simple_tracker_example()

    export_filename = f"test{export_format}"

    # string type path
    fn = os.path.join(tmp_path, export_filename)
    tracker.export(fn, obj_type="obj_type_1")

    # Pathlib type path
    fn = Path(tmp_path) / export_filename
    tracker.export(fn, obj_type="obj_type_1")

    if export_format:
        assert os.path.exists(fn)


@pytest.mark.parametrize("shuffle_objects", [False, True])
def test_write_tracks_only(
    test_real_objects, hdf5_file_path, default_rng, shuffle_objects
):
    """Test writing tracks only using the file handler."""

    if shuffle_objects:
        default_rng.shuffle(test_real_objects)

    # tracker, _ = simple_tracker_example()
    tracker = full_tracker_example(test_real_objects)
    tracks = tracker.tracks

    with btrack.io.HDF5FileHandler(hdf5_file_path, "w") as h:
        h.write_tracks(tracks)

    # now try to read those objects and compare with those used to write
    with btrack.io.HDF5FileHandler(hdf5_file_path, "r") as h:
        tracks_from_file = h.tracks

    for orig, read in zip(tracks, tracks_from_file):
        assert isinstance(orig, btrack.btypes.Tracklet)
        assert isinstance(read, btrack.btypes.Tracklet)

        gt_track = orig.to_dict()
        io_track = read.to_dict()

        for key, gt_value in gt_track.items():
            io_value = io_track[key]
            np.testing.assert_allclose(gt_value, io_value)


def test_write_lbep(tmp_path, test_real_objects):
    """Test writing the LBEP file."""
    tracker = full_tracker_example(test_real_objects)
    tracks = tracker.tracks

    fn = Path(tmp_path) / "LBEP_test.txt"
    btrack.io.export_LBEP(fn, tracker.tracks)

    # check that the file contains the correct number of lines
    with open(fn, "r") as lbep_file:
        entries = lbep_file.readlines()
    assert len(entries) == len(tracks)
    # and that the LBEP entries match
    for entry in entries:
        lbep = [int(e) for e in entry.strip("/n").split()]
        track = next(filter(lambda t: lbep[0] == t.ID, tracks))
        assert lbep == [track.ID, track.start, track.stop, track.parent]


def test_write_hdf_segmentation(hdf5_file_path):
    """Test writing a segmentation to the hdf file."""
    segmentation = np.random.randint(0, 255, size=(100, 64, 64))
    with btrack.io.HDF5FileHandler(hdf5_file_path, "w") as h:
        h.write_segmentation(segmentation)

    with btrack.io.HDF5FileHandler(hdf5_file_path, "r") as h:
        segmentation_from_file = h.segmentation
    np.testing.assert_equal(segmentation, segmentation_from_file)



from __future__ import annotations

import csv
import os
from typing import List

from btrack import btypes


def import_CSV(filename: os.PathLike) -> List[btypes.PyTrackObject]:
    """Import localizations from a CSV file.

    Parameters
    ----------
    filename : PathLike
        The filename of the CSV to import

    Returns
    -------
    objects : List[btypes.PyTrackObject]
        A list of objects in the CSV file.

    Notes
    -----
    CSV file should have one of the following format.

    .. list-table:: CSV header format
       :widths: 20 20 20 20 20
       :header-rows: 1

       * - t
         - x
         - y
         - z
         - label
       * - required
         - required
         - required
         - optional
         - optional
    """

    objects = []
    with open(filename, "r") as csv_file:
        csvreader = csv.DictReader(csv_file, delimiter=",", quotechar="|")
        for i, row in enumerate(csvreader):
            data = {k: float(v) for k, v in row.items()}
            data.update({"ID": i})
            obj = btypes.PyTrackObject.from_dict(data)
            objects.append(obj)
    return objects



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import os

import magicgui
from magicgui.types import FileDialogMode


def save_path_dialogue_box() -> os.PathLike:
    """Helper function to open a save configuration file dialog."""

    app = magicgui.application.use_app()
    show_file_dialog = app.get_obj("show_file_dialog")

    return show_file_dialog(
        mode=FileDialogMode.OPTIONAL_FILE,
        caption="Specify file to save btrack configuration",
        start_path=None,
        filter="*.json",
    )


def load_path_dialogue_box() -> os.PathLike:
    """Helper function to open a load configuration file dialog."""

    app = magicgui.application.use_app()
    show_file_dialog = app.get_obj("show_file_dialog")

    return show_file_dialog(
        mode=FileDialogMode.EXISTING_FILE,
        caption="Choose JSON file containing btrack configuration",
        start_path=None,
        filter="*.json",
    )



import ctypes
import itertools
import logging
import os
import warnings
from typing import List, Optional, Tuple, Union

import numpy as np

from btrack import _version

from . import btypes, config, constants, libwrapper, models, utils
from .io import export_delegator, localizations_to_objects
from .optimise import hypothesis, optimiser

__version__ = _version.version

# get the logger instance
logger = logging.getLogger(__name__)


class BayesianTracker:
    """BayesianTracker.

    BayesianTracker is a multi object tracking algorithm, specifically used to
    reconstruct tracks in crowded fields. Here we use a probabilistic network of
    information to perform the trajectory linking. This class is a wrapper for
    the C++ implementation of the BayesianTracker.

    Parameters
    ----------
    verbose : bool
        A flag to set the verbosity level while logging the output.

    Attributes
    ----------
    n_tracks : int
        The number of found tracks.
    n_dummies : int
        The number of inserted dummy objects.
    tracks : list
        A list of Tracklet objects.
    refs : list
        References to the objects forming the tracks.
    dummies : list
        The dummy objects inserted by the tracker.
    volume : tuple
        The imaging volume as [(xlo, xhi), ..., (zlo, zhi)]. See
        :py:meth:`btrack.btypes.ImagingVolume` for more details.
    frame_range : tuple
        The frame range for tracking, essentially the last dimension of volume.
    LBEP : List[List]
        Return an LBEP table of the track lineages.
    configuration : config.TrackerConfig
        Return the current configuration.

    Notes
    -----
    This method uses positional information (position, velocity ...) as well as
    visual information (labels, features...) for track linking.

    The tracking algorithm assembles reliable sections of track that do not
    contain splitting events (tracklets). Each new tracklet initiates a
    probabilistic model in the form of a Kalman filter [1]_, and utilises this to
    predict future states (and error in states) of each of the objects in the
    field of view.  We assign new observations to the growing tracklets
    (linking) by evaluating the posterior probability of each potential linkage
    from a Bayesian belief matrix for all possible linkages [2]_. The best
    linkages are those with the highest posterior probability.

    Data can be passed in in the following formats:

        * numpy arrays
        * :py:class:`btrack.btypes.PyTrackObject`
        * CSV (see :py:meth:`btrack.io.import_CSV`)
        * HDF (see :py:class:`btrack.io.HDF5FileHandler`)

    The tracker can be used to return all of the original data neatly packaged
    into tracklet objects, or as a nested list of references to the original
    data sets. The latter is useful if using only the first part of a tracking
    protocol, or other metadata is needed for further analysis. The references
    can be used to make symbolic links in HDF5 files, for example. Use
    `optimise` to generate hypotheses for global optimisation [3]_ [4]_. Read the
    :py:class:`optimiser.TrackOptimiser` documentation for more information about the
    track linker.

    Full details of the implementation can be found in [5]_ and [6]_.

    Examples
    --------
    Can be used with ContextManager support, like this:

        >>> with BayesianTracker() as tracker:
        >>>    tracker.configure("./models/cell_config.json")
        >>>    tracker.append(observations)
        >>>    tracker.track()
        >>>    tracks = tracker.tracks

    References
    ----------
    .. [1] 'A new approach to linear filtering and prediction problems.'
      Kalman RE, 1960 Journal of Basic Engineering

    .. [2] 'A Bayesian algorithm for tracking multiple moving objects in outdoor
      surveillance video', Narayana M and Haverkamp D 2007 IEEE

    .. [3] 'Report Automated Cell Lineage Construction' Al-Kofahi et al.
      Cell Cycle 2006 vol. 5 (3) pp. 327-335

    .. [4] 'Reliable cell tracking by global data association', Bise et al.
      2011 IEEE Symposium on Biomedical Imaging pp. 1004-1010

    .. [5] 'Local cellular neighbourhood controls proliferation in cell
      competition', Bove A, Gradeci D, Fujita Y, Banerjee S, Charras G and
      Lowe AR 2017 Mol. Biol. Cell vol 28 pp. 3215-3228

    .. [6] 'Automated deep lineage tree analysis using a Bayesian single cell
      tracking approach', Ulicna K, Vallardi G, Charras G and Lowe AR 2021 Front.
      Comput. Sci. 3
    """

    def __init__(
        self,
        verbose: bool = True,  # noqa: FBT001 FBT002
    ):
        """Initialise the BayesianTracker C++ engine and parameters."""

        # load the library, get an instance of the engine
        self._initialised = False
        self._lib = libwrapper.get_library()
        self._engine = self._lib.new_interface(verbose)

        if not verbose:
            logger.setLevel(logging.WARNING)

        # store a default config
        self._config = config.TrackerConfig(verbose=verbose)

        # silently set the update method to EXACT
        self._lib.set_update_mode(
            self._engine, self.configuration.update_method.value
        )

        # default parameters and space for stored objects
        self._objects: List[btypes.PyTrackObject] = []
        self._frame_range = [0, 0]

    def __enter__(self):
        logger.info("Starting BayesianTracker session")
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        logger.info("Ending BayesianTracker session")
        self._lib.del_interface(self._engine)

    def configure_from_file(self, filename: os.PathLike) -> None:
        """Configure the tracker from a configuration file. See `configure`."""
        warnings.warn(  # noqa: B028
            "This function will be deprecated. Use `.configure()` instead.",
            DeprecationWarning,
        )
        self.configure(filename)

    def configure(
        self, configuration: Union[dict, os.PathLike, config.TrackerConfig]
    ) -> None:
        """Configure the tracker with a motion model, an object model and
        hypothesis generation_parameters.

        Parameters
        ----------
        configuration :
            A dictionary containing the configuration options for a tracking
            session.
        """
        if isinstance(configuration, dict):
            configuration = config.TrackerConfig(**configuration)
        elif isinstance(configuration, (str, os.PathLike)):
            configuration = config.load_config(configuration)

        self._config = configuration

        # set all configuration options using setattr
        for attr in configuration.__fields__:
            setattr(self, attr, getattr(configuration, attr))

        self._initialised = True

    @property
    def configuration(self) -> config.TrackerConfig:
        """Get the current configuration."""
        return self._config

    def __getattr__(self, attr):
        """Default to config if we do not have a specific getter/setter."""
        return getattr(self.configuration, attr)

    def __setattr__(self, attr, value):
        if not attr.startswith("_") and self.configuration.verbose:
            logger.info(f"Setting {attr} -> {value}")

        if attr in config.TrackerConfig.__fields__:
            setattr(self.configuration, attr, value)
        else:
            object.__setattr__(self, attr, value)

        # if we need to update the C++ library instance, do it here
        attrs_to_update = [
            "motion_model",
            "object_model",
            "max_search_radius",
            "volume",
            "update_method",
            "store_candidate_graph",
        ]
        if attr in attrs_to_update and value is not None:
            update_lib_func = getattr(self, f"_{attr}")
            update_lib_func(value)

    def __len__(self) -> int:
        return self.n_tracks

    def _max_search_radius(self, max_search_radius: int):
        """Set the maximum search radius for fast cost updates."""
        self._lib.max_search_radius(self._engine, max_search_radius)

    def _update_method(self, method: Union[str, constants.BayesianUpdates]):
        """Set the method for updates, EXACT, APPROXIMATE, CUDA etc..."""
        self._lib.set_update_mode(self._engine, method.value)

    def _store_candidate_graph(
        self,
        store_graph: bool,  # noqa: FBT001
    ) -> None:
        """Set the flag to store the candidate graph."""
        self._lib.set_store_candidate_graph(self._engine, store_graph)

    @property
    def n_tracks(self) -> int:
        """Return the number of tracks found."""
        return self._lib.size(self._engine)

    @property
    def n_dummies(self) -> int:
        """Return the number of dummy objects (negative ID)."""
        return len(
            [d for d in itertools.chain.from_iterable(self.refs) if d < 0]
        )

    @property
    def tracks(self) -> List[btypes.Tracklet]:
        """Return a sorted list of tracks, default is to sort by increasing
        length."""
        return [self[i] for i in range(self.n_tracks)]

    @property
    def refs(self):
        """Return tracks as a list of IDs (essentially pointers) to the original
        objects. Use this to write out HDF5 tracks."""
        tracks = []
        for i in range(self.n_tracks):
            # get the track length
            n = self._lib.track_length(self._engine, i)

            # set up some space for the output and  get the track data
            refs = np.zeros((n,), dtype=np.int32)
            _ = self._lib.get_refs(self._engine, refs, i)
            tracks.append(refs.tolist())

        return tracks

    @property
    def dummies(self):
        """Return a list of dummy objects."""
        return [
            self._lib.get_dummy(self._engine, -(i + 1))
            for i in range(self.n_dummies)
        ]

    @property
    def LBEP(self):
        """Return an LBEP list describing the track lineage information.

        Notes
        -----

        .. list-table:: LBEP table layout
           :widths: 10 90
           :header-rows: 1

           * - Index
             - Description
           * - L
             - A unique label of the track (label of markers, 16-bit positive).
           * - B
             - A zero-based temporal index of the frame in which the track begins.
           * - E
             - A zero-based temporal index of the frame in which the track ends.
           * - P
             - Label of the parent track (0 is used when no parent is defined).
           * - R
             - Label of the root track.
           * - G
             - Generational depth (from root).
        """
        return utils._lbep_table(self.tracks)

    def _sort(self, tracks: List[btypes.Tracklet]) -> List[btypes.Tracklet]:
        """Return a sorted list of tracks"""
        return sorted(tracks, key=lambda t: len(t), reverse=True)

    def _volume(self, volume: Union[tuple, btypes.ImagingVolume]) -> None:
        """Set the imaging volume.

        Parameters
        ----------
        volume : tuple, ImagingVolume
            A tuple describing the imaging volume.
        """
        volume = btypes.ImagingVolume(*volume)

        # if we've only provided 2 dims, set the last one to a default
        if volume.ndim == constants.Dimensionality.TWO:
            z = (-1e5, 1e5)
            volume = btypes.ImagingVolume(volume.x, volume.y, z)

        self._lib.set_volume(self._engine, np.array(volume, dtype=float))

    def _motion_model(self, model: models.MotionModel) -> None:
        """Set a new motion model. Must be of type MotionModel, either loaded
        from file or instantiating a MotionModel.

        Parameters
        ----------
        model : MotionModel
            A motion model to be used by the tracker.
        """

        # need to populate fields in the C++ library
        self._lib.motion(
            self._engine,
            model.measurements,
            model.states,
            model.A,
            model.H,
            model.P,
            model.Q,
            model.R,
            model.dt,
            model.accuracy,
            model.max_lost,
            model.prob_not_assign,
        )

    def _object_model(self, model: models.ObjectModel) -> None:
        """Set a new object model. Must be of type ObjectModel, either loaded
        from file or instantiating an ObjectModel.

        Parameters
        ----------
        model : ObjectModel
        """
        # if model is not None:
        logger.info(f"Loading object model: {model.name}")
        self.configuration.object_model = model

        # need to populate fields in the C++ library
        self._lib.model(
            self._engine,
            model.states,
            model.emission,
            model.transition,
            model.start,
        )

    @property
    def frame_range(self) -> Tuple[int, int]:
        """Return the frame range."""
        return tuple(self.configuration.frame_range)

    @property
    def objects(self) -> List[btypes.PyTrackObject]:
        """Return the list of objects added through the append method."""
        return self._objects

    def append(
        self, objects: Union[List[btypes.PyTrackObject], np.ndarray]
    ) -> None:
        """Append a single track object, or list of objects to the stack. Note
        that the tracker will automatically order these by frame number, so the
        order here does not matter. This means several datasets can be
        concatenated easily, by running this a few times.

        Parameters
        ----------
        objects : list, np.ndarray
            A list of objects to track.
        """

        objects = localizations_to_objects(objects)

        for idx, obj in enumerate(objects):
            if not isinstance(obj, btypes.PyTrackObject):
                raise TypeError("track_object must be a `PyTrackObject`")

            obj.ID = idx + len(self._objects)  # make sure ID tracks properly

            # call the `obj.set_features` method for each object, which will
            # set the pointer to the array for the tracker
            obj.set_features(self.configuration.features)

            self._frame_range[1] = max(obj.t, self._frame_range[1])
            _ = self._lib.append(self._engine, obj)

        # store a copy of the list of objects
        self._objects += objects

    def _stats(self, info_ptr: ctypes.pointer) -> btypes.PyTrackingInfo:
        """Cast the info pointer back to an object."""

        if not isinstance(info_ptr, ctypes.POINTER(btypes.PyTrackingInfo)):
            raise TypeError("Stats requires the pointer to the object")

        return info_ptr.contents

    def track_interactive(self, *args, **kwargs) -> None:
        logger.warning(
            "`track_interactive` will be deprecated. Use `track` instead."
        )
        return self.track(*args, **kwargs)

    def track(
        self,
        *,
        step_size: int = 100,
        tracking_updates: Optional[
            List[Union[str, constants.BayesianUpdateFeatures]]
        ] = None,
    ) -> None:
        """Run the tracking in an interactive mode.

        Parameters
        ----------
        step_size : int, default=100
            The number of tracking steps to be taken before returning summary
            statistics. The tracking will be followed to completion, regardless
            of the step size provided.
        tracking_updates : list, optional
            A list of tracking updates to perform. See
            :py:class:`btrack.btypes.BayesianUpdateFeatures` for details.
        """

        logger.info("Starting tracking... ")

        if tracking_updates is not None:
            self.configuration.tracking_updates = tracking_updates

        logger.info(
            f"Update using: {[f.name for f in self.configuration.tracking_updates]}"
        )

        # bitwise OR is equivalent to int sum here
        self._lib.set_update_features(
            self._engine,
            sum([int(f.value) for f in self.configuration.tracking_updates]),
        )

        stats = self.step()
        frame = 0

        # while not stats.complete and stats.error not in constants.ERRORS:
        while stats.tracker_active:
            logger.info(
                f"Tracking objects in frames {frame} to "
                f"{min(frame+step_size-1, self._frame_range[1]+1)} "
                f"(of {self._frame_range[1]+1})..."
            )

            stats = self.step(step_size)
            utils.log_stats(stats.to_dict())
            frame += step_size

        if not utils.log_error(stats.error):
            logger.info("SUCCESS.")
            logger.info(
                f" - Found {self.n_tracks} tracks in "
                f"{1+self._frame_range[1]} frames "
                f"(in {stats.t_total_time}s)"
            )
            logger.info(
                f" - Inserted {self.n_dummies} dummy objects to fill "
                "tracking gaps"
            )

    def step(self, n_steps: int = 1) -> Optional[btypes.PyTrackingInfo]:
        """Run an iteration (or more) of the tracking. Mostly for interactive
        mode tracking."""
        if not self._initialised:
            return None
        return self._stats(self._lib.step(self._engine, n_steps))

    def hypotheses(self) -> List[hypothesis.Hypothesis]:
        """Calculate and return hypotheses using the hypothesis engine."""

        if not self.hypothesis_model:
            raise AttributeError("Hypothesis model has not been specified.")

        n_hypotheses = self._lib.create_hypotheses(
            self._engine,
            self.hypothesis_model.as_ctype(),
            self._frame_range[0],
            self._frame_range[1],
        )

        # now get all of the hypotheses
        h = [
            self._lib.get_hypothesis(self._engine, h)
            for h in range(n_hypotheses)
        ]
        return h

    def optimize(self, **kwargs):
        """Proxy for `optimise` for our American friends ;)"""
        return self.optimise(**kwargs)

    def optimise(
        self, options: Optional[dict] = None
    ) -> List[hypothesis.Hypothesis]:
        """Optimize the tracks.

        Parameters
        ----------
        options : dict
            A set of options to be used by GLPK during convex optimization.

        Returns
        -------
        optimized : list
            The list of hypotheses which represents the optimal solution.

        Notes
        -----
        This generates the hypotheses for track merges, branching etc, runs the
        optimiser and then performs track merging, removal of track fragments,
        renumbering and assignment of branches.
        """
        logger.info(f"Loading hypothesis model: {self.hypothesis_model.name}")

        logger.info(
            f"Calculating hypotheses (relax: {self.hypothesis_model.relax})..."
        )
        hypotheses = self.hypotheses()

        # if we have not been provided with optimizer options, use the default
        # from the configuration.
        options = (
            options
            if options is not None
            else self.configuration.optimizer_options
        )

        # if we don't have any hypotheses return
        if not hypotheses:
            logger.warning("No hypotheses could be found.")
            return []

        # set up the track optimiser
        track_linker = optimiser.TrackOptimiser(options=options)
        track_linker.hypotheses = hypotheses
        selected_hypotheses = track_linker.optimise()
        optimised = [hypotheses[i] for i in selected_hypotheses]

        if not optimised:
            logger.warning("Optimization failed.")
            return []

        h_original = [h.hypothesis_type for h in hypotheses]
        h_optimise = [h.hypothesis_type for h in optimised]
        h_types = sorted(set(h_original), key=lambda h: h.value)

        for h_type in h_types:
            logger.info(
                f" - {h_type}: {h_optimise.count(h_type)}"
                f" (of {h_original.count(h_type)})"
            )
        logger.info(f" - TOTAL: {len(hypotheses)} hypotheses")

        # now that we have generated the optimal sequence, merge all of the
        # tracks, delete fragments and assign divisions
        h_array = np.array(selected_hypotheses, dtype=np.uint32)
        h_array = h_array[np.newaxis, ...]
        self._lib.merge(self._engine, h_array, len(selected_hypotheses))
        logger.info(f"Completed optimization with {self.n_tracks} tracks")

        return optimised

    def __getitem__(self, idx: int) -> btypes.Tracklet:
        """Return a single track from the BayesianTracker object."""
        # get the track length
        n = self._lib.track_length(self._engine, idx)

        # set up some space for the output
        children = np.zeros((2,), dtype=np.int32)  # pointers to children
        refs = np.zeros((n,), dtype=np.int32)  # pointers to objects

        # get the track data
        _ = self._lib.get_refs(self._engine, refs, idx)
        nc = self._lib.get_children(self._engine, children, idx)
        p = self._lib.get_parent(self._engine, idx)
        f = constants.Fates(self._lib.get_fate(self._engine, idx))

        # get the track ID
        trk_id = self._lib.get_ID(self._engine, idx)

        # convert the array of children to a python list
        c = children.tolist() if nc > 0 else []

        # now build the track from the references
        refs = refs.tolist()
        dummies = [self._lib.get_dummy(self._engine, d) for d in refs if d < 0]

        track = []
        for r in refs:
            if r < 0:
                # TODO(arl): softmax scores are zero for dummy objects
                dummy = dummies.pop(0)
                dummy.probability = np.zeros((5,), dtype=np.float32)
                track.append(dummy)
            else:
                track.append(self._objects[r])

        # make a new track object and return it
        trk = btypes.Tracklet(trk_id, track, parent=p, children=c, fate=f)
        trk.root = self._lib.get_root(self._engine, idx)
        trk.generation = self._lib.get_generation(self._engine, idx)

        if not self.return_kalman:
            return trk

        # get the size of the Kalman arrays
        sz_mu = self.motion_model.measurements + 1
        sz_cov = self.motion_model.measurements**2 + 1

        # otherwise grab the kalman filter data
        kal_mu = np.zeros((n, sz_mu), dtype=np.float64)  # kalman filtered
        kal_cov = np.zeros((n, sz_cov), dtype=np.float64)  # kalman covariance
        kal_pred = np.zeros((n, sz_mu), dtype=np.float64)  # motion predict

        _ = self._lib.get_kalman_mu(self._engine, kal_mu, idx)
        _ = self._lib.get_kalman_covar(self._engine, kal_cov, idx)
        _ = self._lib.get_kalman_pred(self._engine, kal_pred, idx)

        # cat the data [mu(0),...,mu(n),cov(0,0),...cov(n,n), pred(0),..]
        trk.kalman = np.hstack((kal_mu, kal_cov[:, 1:], kal_pred[:, 1:]))
        return trk

    def export(
        self,
        filename: os.PathLike,
        obj_type: Optional[str] = None,
        filter_by: Optional[str] = None,
    ) -> None:
        """Export tracks using the appropriate exporter.

        Parameters
        ----------
        filename : str
            The filename to export the data. The extension (e.g. .h5) is used
            to select the correct export function.
        obj_type : str, optional
            The object type to export the data. Usually `obj_type_1`
        filter_by : str, optional
            A string that represents how the data has been filtered prior to
            tracking, e.g. using the object property `area>100`
        """
        export_delegator(
            filename, self, obj_type=obj_type, filter_by=filter_by
        )

    def to_napari(
        self,
        replace_nan: bool = True,  # noqa: FBT001,FBT002
        ndim: Optional[int] = None,
    ) -> Tuple[np.ndarray, dict, dict]:
        """Return the data in a format for a napari tracks layer.
        See :py:meth:`btrack.utils.tracks_to_napari`."""

        ndim = self.configuration.volume.ndim if ndim is None else ndim

        return utils.tracks_to_napari(
            self.tracks, ndim=ndim, replace_nan=replace_nan
        )

    def candidate_graph_edges(self) -> List[btypes.PyGraphEdge]:
        """Return the edges from the full candidate graph."""
        num_edges = self._lib.num_edges(self._engine)
        if num_edges < 1:
            logger.warning(
                "No edges were found in the candidate graph. "
                "``config.store_candidate_graph`` is set to "
                f"{self.configuration.store_candidate_graph}"
            )
        return [
            self._lib.get_graph_edge(self._engine, idx)
            for idx in range(num_edges)
        ]



from btrack.napari.reader import get_reader


def test_reader(hdf5_file_path_or_paths):
    reader = get_reader(hdf5_file_path_or_paths)
    assert reader is not None

    tracks = reader(hdf5_file_path_or_paths)
    assert isinstance(tracks, list)
    # TODO: update the HDF file so that it has some tracks that can be read.
    # For now this just checks that the reader can read an empty HDF5 file
    # with no tracks in it
    assert len(tracks) == 0



from .exporters import export_CSV, export_delegator, export_LBEP
from .hdf import HDF5FileHandler
from .importers import import_CSV
from .utils import (
    localizations_to_objects,
    objects_from_array,
    objects_from_dict,
)

__all__ = [
    "export_CSV",
    "export_delegator",
    "export_LBEP",
    "HDF5FileHandler",
    "import_CSV",
    "localizations_to_objects",
    "objects_from_array",
    "objects_from_dict",
]



from __future__ import annotations

import logging
from typing import List, Optional

import numpy as np
from skimage.util import map_array

# import core
from . import btypes, constants
from .constants import DEFAULT_EXPORT_PROPERTIES, Dimensionality
from .io._localization import segmentation_to_objects
from .models import HypothesisModel, MotionModel, ObjectModel

# Choose a subset of classes/functions to document in public facing API
__all__ = [
    "segmentation_to_objects",
    "update_segmentation",
]

# get the logger instance
logger = logging.getLogger(__name__)


def log_error(err_code) -> bool:
    """Take an error code from the tracker and log an error for the user."""
    error = constants.Errors(err_code)
    if (
        error != constants.Errors.SUCCESS
        and error != constants.Errors.NO_ERROR
    ):
        logger.error(f"ERROR: {error}")
        return True
    return False


def log_stats(stats: dict) -> None:
    """Take the statistics from the track and log the output."""

    if log_error(stats["error"]):
        return

    logger.info(
        " - Timing (Bayesian updates: {:.2f}ms, Linking:"
        " {:.2f}ms)".format(stats["t_update_belief"], stats["t_update_link"])
    )

    logger.info(
        " - Probabilities (Link: {:.5f}, Lost:"
        " {:.5f})".format(stats["p_link"], stats["p_lost"])
    )

    if stats["complete"]:
        return

    logger.info(
        " - Stats (Active: {:d}, Lost: {:d}, Conflicts "
        "resolved: {:d})".format(
            stats["n_active"], stats["n_lost"], stats["n_conflicts"]
        )
    )


def read_motion_model(cfg: dict) -> Optional[MotionModel]:
    cfg = cfg.get("MotionModel", {})
    if not cfg:
        return None
    return MotionModel(**cfg)


def read_object_model(cfg: dict) -> Optional[ObjectModel]:
    cfg = cfg.get("ObjectModel", {})
    if not cfg:
        return None
    return ObjectModel(**cfg)


def read_hypothesis_model(cfg: dict) -> Optional[HypothesisModel]:
    cfg = cfg.get("HypothesisModel", {})
    if not cfg:
        return None
    return HypothesisModel(**cfg)


def crop_volume(objects, volume=constants.VOLUME):
    """Return a list of objects that fall within a certain volume."""
    axes = zip(["x", "y", "z", "t"], volume)

    def within(o):
        return all(
            [getattr(o, a) >= v[0] and getattr(o, a) <= v[1] for a, v in axes]
        )

    return [o for o in objects if within(o)]


def _lbep_table(tracks: List[btypes.Tracklet]) -> np.array:
    """Create an LBEP table from a track."""
    return np.asarray([trk.LBEP() for trk in tracks], dtype=np.int32)


def _cat_tracks_as_dict(
    tracks: list[btypes.Tracklet], properties: List[str]
) -> dict:
    """Concatenate all tracks as dictionary."""
    assert all([isinstance(t, btypes.Tracklet) for t in tracks])

    data: dict = {}

    for track in tracks:
        trk = track.to_dict(properties)

        for key in trk:
            trk_property = np.asarray(trk[key])

            # if we have a scalar value, repeat it so the dimensions match
            if trk_property.ndim == 0:
                trk_property = np.repeat(trk_property, len(track))

            if trk_property.ndim > constants.Dimensionality.TWO:
                raise ValueError(
                    f"Track properties of {trk_property.ndim} dimensions are "
                    "not currently supported."
                )

            assert trk_property.shape[0] == len(track)

            if trk_property.ndim == constants.Dimensionality.TWO:
                for idx in range(trk_property.shape[-1]):
                    tmp_key = f"{key}-{idx}"
                    if tmp_key not in data:
                        data[tmp_key] = []
                    data[tmp_key].append(trk_property[..., idx])

            else:
                if key not in data:
                    data[key] = []
                data[key].append(trk_property)

    for key in data:
        data[key] = np.concatenate(data[key])

    return data


def tracks_to_napari(
    tracks: list[btypes.Tracklet],
    *,
    ndim: int | None = None,
    replace_nan: bool = True,
):
    """Convert a list of Tracklets to napari format input.

    Parameters
    ----------
    tracks : list[btypes.Tracklet]
        A list of tracklet objects from BayesianTracker.
    ndim : int or None
        The number of spatial dimensions of the data. If not specified, the
        function attempts to guess the final dimensionality using the z
        coordinates. If specified, it must have a value of 2 or 3.
    replace_nan : bool
        Replace instances of NaN/inf in the track properties with an
        interpolated value.

    Returns
    -------
    data : array (N, D+1)
        Coordinates for N points in D+1 dimensions. ID,T,(Z),Y,X. The first
        axis is the integer ID of the track. D is either 3 or 4 for planar
        or volumetric timeseries respectively.
    properties : dict {str: array (N,)}
        Properties for each point. Each property should be an array of length N,
        where N is the number of points.
    graph : dict {int: list}
        Graph representing associations between tracks. Dictionary defines the
        mapping between a track ID and the parents of the track. This can be
        one (the track has one parent, and the parent has >=1 child) in the
        case of track splitting, or more than one (the track has multiple
        parents, but only one child) in the case of track merging.

    Notes
    -----
    Track properties that are multi-dimensional (>1 dim) will be split according
    to dimension and returned as separate keys. For example, property `softmax`,
    with dimensions (5,) would be split into `softmax-0` ... `softmax-4` for
    representation in napari.
    """
    # guess the dimensionality from the data by checking whether the z values
    # are all zero. If all z are zero then the data are planar, i.e. 2D
    if ndim is None:
        z = np.concatenate([track.z for track in tracks])
        ndim = Dimensionality.THREE if np.any(z) else Dimensionality.TWO

    if ndim not in (Dimensionality.TWO, Dimensionality.THREE):
        raise ValueError("ndim must be 2 or 3 dimensional.")

    t_header = ["ID", "t"] + ["z", "y", "x"][-ndim:]
    p_header = ["t", "state", "generation", "root", "parent"]

    # ensure lexicographic ordering of tracks
    ordered = sorted(tracks, key=lambda t: t.ID)
    header = t_header + p_header
    tracks_as_dict = _cat_tracks_as_dict(ordered, header)

    # note that there may be other metadata in the tracks, grab that too
    prop_keys = p_header + [k for k in tracks_as_dict if k not in t_header]

    # get the data for napari
    data = np.stack(
        [v for k, v in tracks_as_dict.items() if k in t_header], axis=1
    )
    properties = {k: v for k, v in tracks_as_dict.items() if k in prop_keys}

    # replace any NaNs in the properties with an interpolated value
    def nans_idx(x):
        return x.nonzero()[0]

    if replace_nan:
        for k, v in properties.items():
            nans = np.isnan(v)
            v[nans] = np.interp(nans_idx(nans), nans_idx(~nans), v[~nans])
            properties[k] = v

    graph = {t.ID: [t.parent] for t in ordered if not t.is_root}
    return data, properties, graph


def update_segmentation(
    segmentation: np.ndarray,
    tracks: list[btypes.Tracklet],
    *,
    color_by: str = "ID",
) -> np.ndarray:
    """Map tracks back into a masked array.

    Parameters
    ----------
    segmentation : np.array
        Array containing a timeseries of single cell masks. Dimensions should be
        ordered T(Z)YX. Assumes that this is not binary and each object has a unique ID.
    tracks : list[btypes.Tracklet]
        A list of :py:class:`btrack.btypes.Tracklet` objects from BayesianTracker.
    color_by : str, default = "ID"
        A value to recolor the segmentation by.

    Returns
    -------
    relabeled : np.array
        Array containing the same masks as segmentation but relabeled to
        maintain single cell identity over time.

    Notes
    -----
    Useful for recoloring a segmentation by a property such as track ID or
    root tree node. Currently the property must be an integer value, greater
    than zero.

    Example
    -------
    >>> tracked_segmentation = btrack.utils.update_segmentation(
    ...     segmentation, tracks, color_by="ID",
    ... )
    """

    keys = {k: i for i, k in enumerate(DEFAULT_EXPORT_PROPERTIES)}

    coords_arr = np.concatenate(
        [
            track.to_array()[~np.array(track.dummy), : len(keys)].astype(int)
            for track in tracks
        ]
    )

    if color_by not in keys:
        raise ValueError(f"Property ``{color_by}`` not found in track.")

    relabeled = np.zeros_like(segmentation)
    for t, single_segmentation in enumerate(segmentation):
        frame_coords = coords_arr[coords_arr[:, 1] == t]

        xc, yc = frame_coords[:, keys["x"]], frame_coords[:, keys["y"]]
        new_id = frame_coords[:, keys[color_by]]

        if single_segmentation.ndim == constants.Dimensionality.TWO:
            old_id = single_segmentation[yc, xc]
        elif single_segmentation.ndim == constants.Dimensionality.THREE:
            zc = frame_coords[:, keys["z"]]
            old_id = single_segmentation[zc, yc, xc]

        relabeled[t] = map_array(single_segmentation, old_id, new_id) * (
            single_segmentation > 0
        )

    return relabeled
